//
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
//  Produced by:
//  Silicon Grail Corporation
//  710 N. Seward
//  Hollywood, CA
//  USA 90038
//  213-871-9100
//
// NAME: ImCdi.C
//
// COMMENTS:
//

#include <math.h>
#include <UT/UT_Standard.h>
#include "ImTrace.h"
#include "Im.h"
#include "ImTools.h"
#include "ImOpt.h"
#include "ImCdi.h"

/***************************** Error Definitions *****************************
 *
 * Errors consist of a type followed by a printf format string with optional
 * arguments.  In general, the type can be defined as 'E_BASE_IM + E_TYPE_MISC'
 *
 *****************************************************************************/

#define E_CDI_HEADER( name ) E_BASE_IM + E_TYPE_MISC, \
		"Cannot read picture header for 'Chalice Direct Image' file, `%s'!", \
		name

#define E_BAD_MAGIC( name ) E_BASE_IM + E_TYPE_MISC, \
		"Bad magic number in 'Chalice Direct Image' file, `%s'!", name

/************************** Description strings ******************************
 *
 * The className string should match the new class name.  You MUST NOT change
 * the className string once you have decided on it and given the working code
 * to users.  It is used as an identifying tag in various places.
 *
 * The formatName (or English name if you like) is what the programs will
 * display for the user to describe this format at run time.  You can change
 * this at will.
 *
 *****************************************************************************/

const char	*ImCdi::className	  = "ImCdi";
const char	*ImCdi::formatName	  = "Chalice Direct Image";

/******************************* Capabilities ********************************
 *
 * These flags describe what things you format can or cannot do.  The common
 * ones that cover the basic Chalice capabilites are given here.
 * See the file 'Im.h' for more details
 *
 ****************************************************************************/

static const Im::Bitfield cdiCapabilities =	IM_CAP_TYPE_UNSIGNED	|
											IM_CAP_TYPE_FLOAT		|
											IM_CAP_SPACE_MONO		|
											IM_CAP_SPACE_RGB		|
											IM_CAP_SPACE_RGBA		|
											IM_CAP_BPC_8			|
											IM_CAP_BPC_16			|
											IM_CAP_BPC_32;

/*****************************************************************************
 *
 * This bit of code makes sure that the format is added to the list of formats
 * maintained by ImFileMaker.  Although all ImFiles formats are DSO loaded, we
 * recommend that you leave the code as given.  Future changes or ports to
 * other architectures may require this.
 *
 *****************************************************************************/

#ifdef IM_DSO
extern "C" {
UTDSOEXPORT void imCdiMaker( )
{
	static ImCdiMaker *maker = new ImCdiMaker( ImExemplar() );
}	//lint !e529
}
#endif

/****************************** ImCdi Header *********************************
 *
 * A header is common to most filetypes.  We define a dummy one here so that
 * the rest of the code makes sense.  You can accomplish the same effect any
 * way you like.
 *
 *****************************************************************************/

#define CDIMAGIC0	0x43646930		// "Cdi0" magic
#define CDIMAGIC1	0x43646931		// "Cdi1" magic
#define CDIVERSION	1				// Version 1, no intel byte swapping
#define CDIVERSION2	2				// Version 2, NT and SGI
#define SWAPPEDCDIVERSION	16777216
#define SWAPPEDCDIVERSION2  33554432
#define BIGEND_TAG	'M'				// Written in BIG Endian
#define LITEND_TAG	'I'				// Written in Little Endian

//
// Create a header that is always 1024 bytes
//
struct CDIDiskHeader
{
	CDIHeader		h;				// the real header
	unsigned char	reserved[1024 - sizeof(CDIHeader)];
};

#define SWAPSHORT(x) x = swapShort(x)
#define SWAPLONG(x) x = swapLong(x)
void
ImCdi::swapHeader(CDIHeader *head)
{
	SWAPLONG(head->magic[0]);
	SWAPLONG(head->magic[1]);
	SWAPLONG(head->version);
	SWAPLONG(head->width);
	SWAPLONG(head->height);
	SWAPLONG(head->channels);
	SWAPLONG(head->bpc);
}

/****************************** ImCdi Options ********************************
 *
 * You will want to define a list of options ONLY if you actually have options.
 * If you don't have any, just remove this section.  See the ImTiff format
 * for examples. The CDI format has no options at this time
 *
 *****************************************************************************/

// NO OPTIONS

/**************************** ImCdi Constructors *****************************
 *
 * Here we get down to the file open.  The first constructor is for reading,
 * the second is for writing.  The ImFile constructor will make sure that
 * the 'fp' FILE * is set correctly so that you can use it to read from or
 * write to.
 *
 *****************************************************************************/

ImCdi::ImCdi( const char		*fname,
			  FILE				*file,
			  const Uint32		 perm ) : ImFile( fname, file, perm )
{
	myWriteLine = 0;			// not used for reading, but init anyways

	//
	// first we read the header, then fill out the various structures
	// that will describe this file.
	//

	// read the file header
	CDIDiskHeader	header;
	// TBA SYSTEM CALL
	if ( fread( &header, sizeof ( CDIHeader ), 1, fp ) != 1 ) {
		error.V( E_NONE, E_CDI_HEADER( name ) );
		return;
	}

	if ( UTlittleEndian() )
	{
		// Version 1 or Swapped Version 1 file
		if (header.h.version == SWAPPEDCDIVERSION)
			swapHeader(&header.h); //SCHWAPP!
		// Version 2 or Swapped Version 2 file
		else if ( (header.h.version == CDIVERSION2) ||
				  (header.h.version == SWAPPEDCDIVERSION2))
		{
			if (header.h.processor == BIGEND_TAG) // Written in Big Endian
				swapHeader(&header.h); //SCHWAPP!
		}
	}
	else if ( (header.h.version == CDIVERSION2) ||
			  (header.h.version == SWAPPEDCDIVERSION2))
	{
		if (header.h.processor == LITEND_TAG) // Written in Lil Endian
			swapHeader(&header.h); //SCHWAPP!
	}

	// check the magic
	if ( header.h.magic[0] != CDIMAGIC0 && header.h.magic[1] != CDIMAGIC1 )	{
		error.V( E_NONE, E_BAD_MAGIC( name ) );
		return;
	}

	// check the version
	if (( header.h.version != CDIVERSION) &&
		( header.h.version != CDIVERSION2 ) ) {
		error.V( E_NONE, E_BAD_MAGIC( name ) );
		return;
	}

	//
	// ImFiles supports multidimensional files for volume data, etc.
	// Chalice only supports 2 dimensional data at this time.
	//
	ImArray<Im::Uint>	dimBounds(2);			// bounds in pixels

	dimBounds[0]		= header.h.width;		// X pixels
	dimBounds[1]		= header.h.height;		// Y pixels

	Uint	numChannels = header.h.channels;	// number of channels

	//
	// This is used to return two attributes to the calling program,
	// 1. IM_PLANAR or IM_INTERLEAVED
	// 2. IM_COMPRESSED
	//
	// Both are used for information only at this time, it does not affect
	// how the data is read.
	//
	Uint	attributes	= IM_INTERLEAVED;		// no compression, interleaved

	//
	// The color space must match the number of channels.  We are assuming
	// that the image is composed of RGB data.  Chalice does not support
	// CMYK or other color spaces at this time, even though ImFiles does.
	//
	ColorSpace colorSpace;
	switch ( numChannels )
	{
	case 1:
		colorSpace = ImMono;			// 1 channels is Mono
		break;

	case 2:
		colorSpace = ImComplex;			// 2 channels is complex
		break;

	case 3:
		colorSpace = ImRGB;				// 3 channels is RGB
		break;

	case 4:
	default:
		colorSpace = ImRGBA;			// 4 or more channels is RGBA
		break;
	}

	//
	// Chalice only supports Unsigned and Float data types at this time.
	// ImFiles supports other types, but we will not cover them here.
	//
	DataType	dataType;
	if ( header.h.bpc < 4 )
		dataType = Unsigned;	// 8 or 16 bit unsigned data
	else
		dataType = Float;		// 32 bit float

	ImArray<Uint>	channelBits(numChannels);		// # of bits per channel
	ImArray<Uint>	channelSigBits(numChannels);	// # of significant bits

	//
	// The bits for each channel are the same for most images.
	// The difference between bits and significant bits is that it allows
	// us to indicate 12 bit data in 16 bit words (for instance).
	// Chalice always read data in the # of sig bits that the format indicates,
	// and the user is responsible for scaling the data inside of Chalice.
	// ImFiles does have the capability of asking for the data to be scaled on
	// read, but Chalice does NOT do that.
	//
	channelBits		= header.h.bpc * 8;
	channelSigBits	= header.h.bpc * 8;

	//
	// This is used for color mapped files.  For non-mapped files, we just
	// need a dummy to pass to ImSpecs.  See ImTiff for a model of reading
	// color mapped files.
	//
	ArrayChannel	channelMaps;					// look-up tables

	//
	// build the file description and add it to the list.  ImFiles supports
	// multi-image files, but Chalice only reads the first image.  The
	// exception are motion files.  They are assumed to be all of the
	// same specifications, so only one ImSpecs object needs to be created
	// in any case
	//
	ImSpecs *d  = new ImSpecs( 2, dimBounds, dataType, numChannels,
							   channelBits, channelSigBits, colorSpace,
							   channelMaps, attributes );
	(void)add( d, After );
	(void)last( );
}

ImCdi::ImCdi(
	const char			*fname,
	FILE				*file,
	const Uint32		 perm,
	const Uint			 numDims,			// # of dimensions
	const ImArray<Uint>	&dimBounds,			// bounds in each dimension
	const Uint			 numChannels,		// # of channels
	const ImArray<Uint>	&bitsPerChannel,	// # of bits per channel
	const ImArray<Uint>	&significantBits,	// # of significant bits per channel
	const DataType		 channelType,		// data type
	const ColorSpace	 colorSpace,		// Color space
	const ArrayChannel	&channelMaps,		// channel lookup tables
	const Uint			 storageType		// compressed, other
	) : ImFile( fname, file, perm, numDims, dimBounds, numChannels,
				bitsPerChannel, significantBits, channelType, colorSpace,
				channelMaps, storageType )
{
	//
	// The ImSpec for this new file has already been created by ImFile
	// we just need to create the header for the file and write it out.
	//

	//
	// You will want to do some checking just to make sure that
	// the parameters passed in are ones you can handle.  Below
	// are some examples.  The error defines are in ImError.h
	//
	if ( numDims != 2 ) {
		error.V(E_NONE, E_DIMS("2", formatName));
		return;
	}

	if ( channelType != Unsigned && channelType != Float ) {
		error.V(E_NONE, E_CHANNEL_TYPE("unsigned or float", formatName));
		return;
	}

	if ( !(bitsPerChannel==8 || bitsPerChannel==16 || bitsPerChannel==32) ||
		 !(significantBits==8 || significantBits==16 || significantBits==32)) {
		error.V(E_NONE, E_BITS("8, 16 or 32", formatName));
		return;
	}

	if ( colorSpace != ImMono && colorSpace != ImRGB && colorSpace != ImRGBA &&
		 colorSpace != ImComplex ) {
		error.V(E_NONE, E_COLORSPACE("Mono, Complex, RGB or RGBA", formatName));
		return;
	}

	//
	// Well, the parameters all seem to be in order, so now we fill out
	// the header and write it.
	//
	CDIDiskHeader	header;

	header.h.magic[0]	= CDIMAGIC0;
	header.h.magic[1]	= CDIMAGIC1;
	header.h.version	= CDIVERSION2;
	header.h.width		= dimBounds[0];
	header.h.height		= dimBounds[1];
	header.h.channels	= numChannels;
	header.h.bpc		= bitsPerChannel[0] / 8;

	if ( UTlittleEndian() )
		header.h.processor = LITEND_TAG;
	else
		header.h.processor = BIGEND_TAG;

	// TBA SYSTEM CALL
	if ( fwrite( &header, sizeof(CDIDiskHeader), 1, fp ) != 1 ) {
		 error.V( E_NONE, E_WRITE_FILE( formatName, name ) );
		 return;
	}

	//
	// Use this to make sure we always write in order
	//
	myWriteLine = 0;
}

//
// Take care of cleaning up any data structures here, as normal.  However
// if something to be done can cause an error, it should be done in cleanUp,
// because then the error can be reported.
//
ImCdi::~ImCdi( )
{
}

//
// Take care of any final processing that can cause an error.  No reads or
// writes will be done after this is called.
//
ImBooln
ImCdi::cleanUp()
{
	return ImTrue;
}

//
// Seek to the specified x, y location in the file
//
ImBooln
ImCdi::iseek( Uint x, Uint y )
{
	Uint	bpp = (bits()[0] / 8) * channels();
	// TBA SYSTEM CALL
	if ( fseek( fp, sizeof(CDIDiskHeader) + y * bounds()[0] * bpp + x * bpp,
				SEEK_SET ) != 0 ) {
		error.V( E_NONE, E_READ_FILE( "CDI", name ) );
		return ImFalse;
	}

	return ImTrue;
}

//
// Read the sub area described by area into buffer.  Buffer will contain
// enough bytes to hold the entire area.  You can make the following
// assumptions when writing Chalice plugins:
//
// 1. Chalice always asks for all channels in RGBAO order.
// 2. Chalice always asks for data at the same bit depth that is in the file.
// 3. Chalice always expects data to be in the lower N bits of the correctly
//    sized word.  EG.  12 bit data should be put in the lower 12 bits of a
//	  16 bit unsigned short.
//
ImBooln
ImCdi::read( const SubArea &area, void *buffer )
{
	//
	// see if we can just read
	//
	if ( area.bounds[0] == bounds()[0] ) {

		if ( ! iseek( 0, area.origin[1] ) )	// seek to start of data
			return ImFalse;		// Failure

		if ( fread( buffer, area.bytesPer(), 1, fp ) != 1 ) {
			error.V( E_NONE, E_READ_FILE( "CDI", name ) );
			return ImFalse;	// Failure
		}
	}
	else {
		//
		// read a line at a time, because we need partial lines
		//
		Uint	 offx	= area.origin[0];				// x start
		Uint	 line	= area.origin[1];				// y start
		Uint	 end	= line + area.bounds[1] - 1;	// y end
		Uint	 bpl	= area.bytesPerPixel() * area.bounds[0]; // bytes/line
		Uchar	*bp		= (Uchar *)buffer;

		while ( line <= end ) {
			if ( ! iseek( offx, line ) )
				return ImFalse;		// Failure
			if ( fread( bp, bpl, 1, fp ) != 1 ) {
				error.V( E_NONE, E_READ_FILE( "CDI", name ) );
				return ImFalse;		// Failure
			}
			bp += bpl;				// move on to the next line
			line++;
		}
	}

	return ImTrue;	// Success!
}

//
// Write the sub area described by outArea into the file.  The buffer will
// contain all pixels described by the outArea.  For Chalice plugins, you
// can safely assume the following:
//
// 1. The first SubArea can be ignored.
// 2. All channels will always be written in RGBAO order.
// 3. Full strips will always be written, eg. outArea.bounds[0] == bounds()[0]
// 4. Data will always be written top to bottom unless the IM_CAP_WRITE_REVERSE
//	  is in the capabilities, in which case the write will be bottom to top,
//	  one line at a time.
// 5. If IM_CAP_MUST_WRITE_PLANAR is in the capabilites, the entire image will
//	  be written at one time, one plane at a time.
//
ImBooln
ImCdi::write( const SubArea &, void *buffer, const SubArea &outArea )
{
	UT_ASSERT( outArea.bounds[0] == bounds()[0] );		// write of full lines
	UT_ASSERT( outArea.origin[1] == myWriteLine );		// write in order

	if ( fwrite( buffer, outArea.bytesPer(), 1, fp ) != 1 ) {
		error.V( E_NONE, E_WRITE_FILE( "CDI", name ) );
		return ImFalse;		// Failure
	}

	//
	// Keep track of the last line written
	//
	myWriteLine += outArea.bounds[1];

	return ImTrue;	// Success!
}

const char	*ImCdi::getClassName() const { return className; }
const char	*ImCdi::getFormatName() const { return formatName; }

/*************************** ImCdiMaker *************************************
 *
 * This class acts as a sort of virtual constructor.  It has an open
 * file method for both read and write, plus some extra helper functions.
 *
 ****************************************************************************/

ImCdiMaker::ImCdiMaker( ImExemplar e ) :
		ImFileMaker( e,						// exemplar magic, just do it
					 ImCdi::className,		// class & format names
					 ImCdi::formatName,
					 "cdi",					// filename extension
					 cdiCapabilities,		// capabilities
					 8600,					// matching confidence
					 0 )					// options
{
}

ImCdiMaker::~ImCdiMaker()
{
}

//
// Create a new Im object that can read this filetype.
//
// name - is the name of the file.  (Can be NULL)
// file - is opened for reading.	(Cannot be NULL)
// perm - should be IM_FILE_READ
// noSearch - if TRUE, assume that this file is of your type and open it.
//			  if FALSE, check the file before trying to create a new object.
//
Im *
ImCdiMaker::openIt( const char		*name,
					FILE			*file,
					Uint32			 perm,
					ImBooln			 noSearch )
{
	CDIHeader	 header;

	if ( noSearch || getHeader( file, header ) ) {
		if ( NOTNULL( file ) )
			rewind( file );

		return new ImCdi( name, file, perm );
	}

	return NULL;
}

//
// Open a file for write/update.
// name	- The name of the file to open.
// file - An already opened file descriptor, the type must match
//		the permission parameter. name or file MUST be non-null.
//		If name and file are set, name is assumed to be pointing to the
//		already opened file.
// permission - is the or of the following masks:
//				IM_FILE_READ
//				IM_FILE_WRITE
//				IM_FILE_CREATE
//
// bounds - The dimensions in each axis, the bounds array
//			must have at least two elements.  Currently,
//			no file format supports more than two, but volume
//			data is supported at a low level.
// numChannels - The number of channels in the new image.
//				Mono images have 1 channel, RGB images have 3 channels, etc.
// bits - The number of bits in the written data, usually 8, 16 or 32.
//		  There must be one entry in bits for every channel.
// sigBits - The significant bits in the output data.  all entries
//			must be <= bits.  There must be one entry for every channel.
// datatype - The type of data being written to the file.
// space - The color space to be written.
// maps - Colormaps to be written.
//
Im *
ImCdiMaker::openIt( const char					*name,
					 FILE						*file,
				     Uint32						 permission,
				     Im::Uint					 numDims,
				     const ImArray<Im::Uint>	&bounds,
				     Im::Uint					 numChannels,
				     const ImArray<Im::Uint>	&bits,
				     const ImArray<Im::Uint>	&sigBits,
				     Im::DataType				 channelType,
				     Im::ColorSpace				 space,
				     const Im::ArrayChannel		&maps,
				     Im::Uint					 attributes )
{
	return new ImCdi( name, file, permission, numDims, bounds, numChannels,
					   bits, sigBits, channelType, space, maps, attributes );
}

//
// This method should return TRUE if file is of your type.
// name - is for your convience only.  It may be ignored.
// file - may be assumed to be opened correctly for reading.
// 		  You should rewind(file) before reading.
// statinfo - the results of a stat on the file.
//
// width, height, channels, bits and type correspond to the
// information given above.  Each one MAY be NULL, so check
// each one before setting it.
//
ImBooln
ImCdiMaker::getFormat(const char		* /*name*/,
					  FILE				*fp,
					  const struct stat	* /*statinfo*/,
					  Im::Uint			*width,
					  Im::Uint			*height,
					  Im::Uint			*channels,
					  Im::Uint			*bits,
					  Im::DataType		*type,
					  Im::Bitfield		*caps)
{
	CDIHeader	 header;

	if ( getHeader( fp, header ) == ImTrue ) {

		if ( width )	*width = header.width;
		if ( height )	*height = header.height;
		if ( channels )	*channels = header.channels;
		if ( bits )		*bits = header.bpc * 8;
		if ( type )		*type = header.bpc == 4 ? Im::Float : Im::Unsigned;
		if ( caps )		*caps = cdiCapabilities;

		return ImTrue;
	}

	return ImFalse;
}

//
// The basic calculation is simple, just add in the size of your header.
// Most formats estimate the size based on uncompressed data, however if
// you can come up with an estimate that is fairly accurate for compressed
// data, feel free to use that.
//
Im::Uint
ImCdiMaker::estimateSize(const ImArray<Im::Uint>	&bounds,
						 const Im::Uint				 numChannels,
						 const ImArray<Im::Uint>	&bits
						 ) const
{
	return sizeof(CDIDiskHeader) + bounds[0] * numChannels * (bits[0]/8) *
		bounds[1];
}

//
// This method is optional.  If the only extension that the format supports
// is the same as the one passed into ImCdiMaker::ImCdiMaker, the base
// class will take care of the extension matching for you.  If you support
// multiple extensions, then you should define this method.
//
#if 0
ImBooln
ImCdiMaker::isExtension( const char *ext )
{
	return ImBooln( !strcasecmp( ext, "cdi" ) );
}
#endif

//
// This method is not required, but it is a method used in all existing
// GWH&A file format readers, so it may prove useful. Notice that the fp
// is not assumed to be rewound
//
ImBooln
ImCdiMaker::getHeader( FILE *fp, CDIHeader &header )
{
	CDIDiskHeader	dh;

	rewind( fp );

	//
	// read the file header
	//
	if ( fread( &dh, sizeof ( CDIDiskHeader ), 1, fp ) != 1 )
		 return ImFalse;

	if ( UTlittleEndian() )
	{
		// Version 1 or Swapped Version 1 file
		if (dh.h.version == SWAPPEDCDIVERSION)
			ImCdi::swapHeader(&dh.h); //SCHWAPP!
		// Version 2 or Swapped Version 2 file
		else if ( (dh.h.version == CDIVERSION2) ||
				  (dh.h.version == SWAPPEDCDIVERSION2))
		{
			if (dh.h.processor == BIGEND_TAG) // Written in Big Endian
				ImCdi::swapHeader(&dh.h); //SCHWAPP!
		}
	}
	else if ( (dh.h.version == CDIVERSION2) ||
			  (dh.h.version == SWAPPEDCDIVERSION2))
	{
		if (dh.h.processor == LITEND_TAG) // Written in Lil Endian
			ImCdi::swapHeader(&dh.h); //SCHWAPP!
	}

	if ( dh.h.magic[0] != CDIMAGIC0   || dh.h.magic[1] != CDIMAGIC1 ||
		 (dh.h.version  != CDIVERSION && dh.h.version != CDIVERSION2) )
		return ImFalse;					// bad magic

	header = dh.h;

	return ImTrue;						// it's a 'Chalice Direct Image' file
}
