//
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Kimball Thurston
//  Silicon Grail Corporation
//  710 N. Seward
//  Hollywood, CA
//  USA 90038
//  213-871-9100
//
// NAME:	ImFileMaker
//
// COMMENTS:
//

#ifndef _ImFileMaker_h_
#define _ImFileMaker_h_

class UT_FileName;

//
// These defines provide a simpler, shorter way to access option arrays.
// They assume you have named your option array 'theOptions'
//
#define ImGETTGL(option) ImFileMaker::getTglOpt(&theOptions[option])
#define ImGETINT(option) ImFileMaker::getIntOpt(&theOptions[option])
#define ImGETFLT(option) ImFileMaker::getFltOpt(&theOptions[option])
#define ImGETSTR(option) ImFileMaker::getStrOpt(&theOptions[option])

//
// This class acts as a 'base' to allow the ImFileMaker class to
// keep a list of formats
//
class ImExemplar {
public:
	ImExemplar() {}
};

//
// Keep a group of options stowed and setable
//
class ImFileOptStow {
public:
	friend class ImFileMaker;

	ImFileOptStow( );
	~ImFileOptStow( );
	void				 setOption(const ImOpt *option, Uint32 toggle);
	void				 setOption(const ImOpt *option, const int *ints);
	void				 setOption(const ImOpt *option, const float *floats);
	void				 setOption(const ImOpt *option, const char *string);
	inline Uint32		 getTglOpt(const ImOpt *option);
	inline const int	*getIntOpt(const ImOpt *option);
	inline const float	*getFltOpt(const ImOpt *option);
	inline const char	*getStrOpt(const ImOpt *option);

	void				 dump(FILE *fp);	// for debugging

private:
	inline Im::Uchar	*getDataPtr(const ImOpt	*option);
	Im::Uchar			*myStowed;
};

//
// This is handy for finding out if a cap is set
//
#define IM_HAS_CAP(bitfield, cap) ((bitfield & cap) == cap)

//
// The ImFileMaker class is used to determine file types and open
// files in a format non-specific way.
//
class ImFileMaker {
public:
	//
	// The constructor for the file maker class needs the following:
	//
	// formatClass		- Used for internal naming
	// formatName		- Used for user readable naming
	// formatExtension	- The default "." extension.
	// capabilities		- The | of the IM_CAP information
	// priority			- how sure you are that this format can be matched
	//
	// optDescription	- a description in ImOpt form of the optData
	// optData			- pointer to the data structure that contains
	//					  format specific options.
	// optSize			- size of the optData structure in bytes.
	//
						 ImFileMaker( ImExemplar	 exemplar,
									  const char	*formatClass,
									  const char	*formatName,
									  const char	*formatExtension,
									  Im::Bitfield	 capabilties,
									  Im::Uint		 priority,
									  ImOpt			*optDescription	= NULL);

	virtual				~ImFileMaker();

	//
	// exemplar routines that open a Im object for a file to be read.
	// Version 1 searches for a format that can open the file
	// Version 2 opens the file given a specific format
	//
	static Im			*open(const char		*name,
							  FILE				*file,
							  const Uint32		 permission );

	static Im			*open(const char		*name,
							  FILE				*file,
							  const char		*formatClass,
							  const Uint32		 permission );

	//
	// exemplar routine that creates a new file with the specified
	// respects
	//
	static Im			*open( const char				*name,
							   FILE						*file,
							   const char				*formatClass,
							   const Uint32				 permission,
							   Im::Uint					 numDims,
							   const ImArray<Im::Uint>	&bounds,
							   Im::Uint					 numChannels,
							   const ImArray<Im::Uint>	&bits,
							   const ImArray<Im::Uint>	&sigBits,
							   Im::DataType				 dataType,
							   Im::ColorSpace			 space,
							   const Im::ArrayChannel	&maps,
							   Im::Uint					 attributes );

	//
	// Determine the file format of a file and return optional information
	// about it.
	//
	static const char	*format(const char		*name,
								ImBooln			 reportErrors,
								Im::Uint		*width	  = NULL,
								Im::Uint		*height	  = NULL,
								Im::Uint		*channels = NULL,
								Im::Uint		*bits	  = NULL,
								Im::DataType	*type	  = NULL,
								Im::Uint		*size	  = NULL,
								Im::Bitfield	*caps	  = NULL);

	//
	// Get information about a file given its format
	//
	static ImBooln		 info(const char	*name,
							  const char	*format,
							  ImBooln		 reportErrors,
							  Im::Uint		*width	  = NULL,
							  Im::Uint		*height	  = NULL,
							  Im::Uint		*channels = NULL,
							  Im::Uint		*bits	  = NULL,
							  Im::DataType	*type	  = NULL,
							  Im::Uint		*size	  = NULL,
							  Im::Bitfield	*caps	  = NULL);

	//
	// Compute the estimated size of the file in bytes
	//
	static Im::Uint		 size( const char				*formatClass,
							   const ImArray<Im::Uint>	&bounds,
							   const Im::Uint			 numChannels,
							   const ImArray<Im::Uint>	&bits );

	//
	// this function determines what capabilities a given file format has
	//
	static Im::Bitfield	 formatGetCap( const char *formatClass );

	//
	// See if the given format is loaded
	//
	static ImBooln		 hasFormat(const char *className);

	//
	// Retrieve the format names, in priority or name sorted order
	//
	static ImBooln		 formatNames(const char	**formatClass,
									 const char	**formatName,
									 const char **formatExtension = 0);

	static ImBooln		 formatNamesSorted(const char	**formatClass,
										   const char	**formatName,
										   const char	**formatExtension = 0);

#ifdef IM_DSO
#pragma TODO( KDT, "This needs to become start and stop..." )
	static void			 loadFormats( );
	static void			 cleanupFormats( );
	static void			 setupPointers( void *stwsize,
										void *stow,
										void *curstow,
										void *xmplr,
										void *lst );
#endif

	//
	// Options handling.  The class ImOpt describes the various option types.
	//

	//
	// getOptionsDescription -	returns pointer to an array of options for
	//							a particular class. The last option has a
	//							type of ImOpt::tNone.
	// setOption -	These four functions set the various options types.
	//				You must give the class, option and values.
	// setOptionsDefaults - Set all options to defaults.
	// stowOptions - returns a pointer to a copy of all the options for all the
	//				 formats.  The first time stowOptions is called it should
	//				 be called with a NULL pointer.  After this, you can re-use
	//				 the pre-allocated buffer.
	// unstowOptions -	Takes a pointer from stow options and sets all the
	//					options according to the data in that buffer.
	// freeOptions - frees the stowed buffer.
	//
	static const ImOpt	*getOptionsDescription(const char *formatClass);
	static void			 initOptions( );
	static void			 initOptions(ImFileOptStow *stowed);
	static void			 setOptionsDefaults( );
	static void			 setOptionsDefaults(ImFileOptStow *stowed);
	static void			 setCurrentOptions(ImFileOptStow *stowed);
	static void			 resetCurrentOptions( );
	inline static void			 setOption(const ImOpt *, Uint32 toggle);
	inline static void			 setOption(const ImOpt *, const int *ints);
	inline static void			 setOption(const ImOpt *, const float *floats);
	inline static void			 setOption(const ImOpt *, const char *string);
	inline static Uint32		 getTglOpt(const ImOpt *option);
	inline static const int		*getIntOpt(const ImOpt *option);
	inline static const float	*getFltOpt(const ImOpt *option);
	inline static const char	*getStrOpt(const ImOpt *option);

	static void			 initStow();

	inline static void	 dumpStow(FILE *fp);

protected:
	//
	// Open a file for read/update.
	// name	- the name of the file to open.
	// file - an already opened file descriptor, the type must match
	//		the permission parameter.
	//		name or file MUST be non-null.  If name and file are set,
	//		name is assumed to be pointing to the already opened file.
	// permission - is the or of the following masks:
	//				IM_FILE_READ
	//				IM_FILE_WRITE
	// noSearch - The call is telling you that the file is of you type,
	//			  don't bother with a header check, just open the file.
	virtual Im			*openIt( const char		*name,
								 FILE			*file,
								 Uint32			 permission,
								 ImBooln		 noSearch );

	//
	// Open a file for write/update.
	// name	- The name of the file to open.
	// file - An already opened file descriptor, the type must match
	//		the permission parameter. name or file MUST be non-null.
	//		If name and file are set, name is assumed to be pointing to the
	//		already opened file.
	// permission - is the or of the following masks:
	//				IM_FILE_READ
	//				IM_FILE_WRITE
	//				IM_FILE_CREATE
	//
	// bounds - The dimensions in each axis, the bounds array
	//			must have at least two elements.  Currently,
	//			no file format supports more than two, but volume
	//			data is supported at a low level.
	// numChannels - The number of channels in the new image.
	//				Mono images have 1 channel, RGB images have 3 channels, etc.
	// bits - The number of bits in the written data, usually 8, 16 or 32.
	//		  There must be one entry in bits for every channel.
	// sigBits - The significant bits in the output data.  all entries
	//			must be <= bits.  There must be one entry for every channel.
	// datatype - The type of data being written to the file.
	// space - The color space to be written.
	// maps - Colormaps to be written.
	//
	virtual Im			*openIt( const char					*name,
								 FILE						*file,
								 Uint32						 permission,
								 Im::Uint					 numDims,
								 const ImArray<Im::Uint>	&bounds,
								 Im::Uint					 numChannels,
								 const ImArray<Im::Uint>	&bits,
								 const ImArray<Im::Uint>	&sigBits,
								 Im::DataType				 dataType,
								 Im::ColorSpace				 space,
								 const Im::ArrayChannel		&maps,
								 Im::Uint					 attributes );

	//
	// This method should return TRUE if file is of your type.
	// name - is for your convience only.  It may be ignored.
	// file - may be assumed to be opened correctly for reading.
	// 		  You should rewind(file) before reading.
	// statinfo - the results of a stat on the file.
	//
	// width, height, channels, bits and type correspond to the
	// information given above.  Each one MAY be NULL, so check
	// each one before setting it.
	//
	virtual ImBooln		 getFormat(const char			*name,
								   FILE					*file,
								   const struct stat	*statinfo,
								   Im::Uint				*width,
								   Im::Uint				*height,
								   Im::Uint				*channels,
								   Im::Uint				*bits,
								   Im::DataType			*type,
								   Im::Bitfield			*caps);

	//
	// Return an estimate of the size of the resulting file based on
	// the input bounds, numChannels and bits.  It is understood that
	// estimates are based on non-compressed data and may therefore
	// be highly inaccurate.
	//
	virtual Im::Uint	 estimateSize( const ImArray<Im::Uint>	&bounds,
									   const Im::Uint			 numChannels,
									   const ImArray<Im::Uint>	&bits
									   ) const;

	//
	// return TRUE if ext is a string that matches you extension.
	// it will be of the form "pic" or "TIFF" or "Rgb".  You should
	// use strcasecmp in most instances.
	//
	virtual ImBooln		 isExtension( const char *ext );

	const char			*getFormatClass( ) { return fileFormatClass; }

	//
	// Convert a ImFiles perm to a FILE * string for the fopen call.
	//
	static char			*permStr( const char * const name_,
								  const Uint32		 perm );

	static Error		error;				// error stack access

private:
	friend				 class ImFileOptStow;

	//
	// return True if the caps say that this is a device
	//
	inline ImBooln		 isDevice( );

	//
	// return True if the caps say that this is something we should format scan
	//
	inline ImBooln		 isFormatScan( );

	//
	// Error message printer for functions that should be implemented by
	// a sub-class
	//
	static void			 SubclassResponsibility(const char *cn, const char *mn);

#ifdef IM_DSO
	//
	// Handle DSO loading, etc.
	//
	static void			 loadFormat( UT_FileName *fmt );
#endif

	//
	// Data from constructor
	//
	// formatClass		- Used for internal naming
	// formatName		- Used for user readable naming
	// formatExtension	- The default "." extension.
	// capabilities		- The | of the IM_CAP information
	// priority			- how sure you are that this format can be matched
	//
	// optDescription	- a description in ImOpt form of the optData
	// optData			- pointer to the data structure that contains
	//					  format specific options.
	// optSize			- size of the optData structure in bytes.
	//
	const char			*fileFormatClass;
	const char			*fileFormatName;
	const char			*fileFormatExtension;
	Im::Bitfield		 capabilities;
	Im::Uint			 priority;
	const ImOpt			*optionsDescription;
	Im::Uint			 optionsSize;
	Im::Uint			 optionsOffset;

	//
	// List management
	//
	ImFileMaker				*next;

	static ImFileMaker		**list;				// list of all loaded formats
	static ImFileMaker		**exemplar;			// the exemplar
	static Im::Uint			*myOptionsStowSize;	// total size of all options
	static ImFileOptStow	**myStow;			// local options
	static ImFileOptStow	**myCurrentStow;	// current options
};	/* end class ImMaker */

inline Im::Uchar *
ImFileOptStow::getDataPtr(const ImOpt	*option)
{
	return myStowed + option->myMaker->optionsOffset + option->getOffset();
}

inline Uint32
ImFileOptStow::getTglOpt(const ImOpt	*opt)
{
	return *(Uint32 *)getDataPtr(opt);
}

inline const int *
ImFileOptStow::getIntOpt(const ImOpt	*opt)
{
	return (int *)getDataPtr(opt);
}

inline const float *
ImFileOptStow::getFltOpt(const ImOpt	*opt)
{
	return (float *)getDataPtr(opt);
}

inline const char *
ImFileOptStow::getStrOpt(const ImOpt	*opt)
{
	return *(char **)getDataPtr(opt);
}

inline void
ImFileMaker::setOption(const ImOpt *option, Uint32 toggle)
{
	(*myCurrentStow)->setOption(option, toggle);
}

inline void
ImFileMaker::setOption(const ImOpt *option, const int *ints)
{
	(*myCurrentStow)->setOption(option, ints);
}

inline void
ImFileMaker::setOption(const ImOpt *option, const float *floats)
{
	(*myCurrentStow)->setOption(option, floats);
}

inline void
ImFileMaker::setOption(const ImOpt *option, const char *string)
{
	(*myCurrentStow)->setOption(option, string);
}

inline Uint32
ImFileMaker::getTglOpt(const ImOpt *option)
{
	return (*myCurrentStow)->getTglOpt(option);
}

inline const int *
ImFileMaker::getIntOpt(const ImOpt *option)
{
	return (*myCurrentStow)->getIntOpt(option);
}

inline const float *
ImFileMaker::getFltOpt(const ImOpt *option)
{
	return (*myCurrentStow)->getFltOpt(option);
}

inline const char *
ImFileMaker::getStrOpt(const ImOpt *option)
{
	return (*myCurrentStow)->getStrOpt(option);
}

inline ImBooln
ImFileMaker::isDevice( )
{
	return ImBooln(capabilities & IM_CAP_DEVICE);
}

inline ImBooln
ImFileMaker::isFormatScan( )
{
	return ImBooln(0 == (capabilities & IM_CAP_NO_FORMAT_SCAN));
}

inline void
ImFileMaker::dumpStow( FILE *fp )
{
	(*myCurrentStow)->dump(fp);
}

#endif
