//
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
//  Produced by:
//  Silicon Grail Corporation
//  710 N. Seward
//  Hollywood, CA
//  USA 90038
//  213-871-9100
//
// NAME: ImTiff.C
//
// COMMENTS:
//
#include <math.h>
#include <unistd.h>
#include <UT/UT_Standard.h>
#include <UT/UT_MathFunc.h>
#include <UT/UT_Endian.h>
#include "ImTools.h"
#include "ImTrace.h"
#include "ImTiff.h"
#include "ImCopy.h"


#define E_TIFF_MISSING(thetag) E_BASE_IM + E_TYPE_TIFF_LIB, \
		"TIFF file %s is missing tag %s", name, thetag

#define E_TIFF_SET(thetag) E_BASE_IM + E_TYPE_TIFF_LIB, \
		"Can't set TIFF tag %s in file %s", thetag, name

#define E_TIFF_ERROR(str) E_BASE_IM + E_TYPE_TIFF_LIB, \
		str

#define E_TIFF_BAD(what) E_BASE_IM + E_TYPE_TIFF_LIB, \
		"Could not complete %s in TIFF file %s. (Bad file?)", what, name

#define E_HANDLER_BEGIN errorCurrent = this; \
				TIFFErrorHandler eHandler = TIFFSetErrorHandler(errorHandler)

#define E_HANDLER_END	(void)TIFFSetErrorHandler(eHandler);

//
// Do a handler end, but make SURE that we put an error message up
//
#define E_HANDLER_END_ERROR(what)			\
{											\
	if ( !error.Occurred() )				\
		error.V(E_NONE, E_TIFF_BAD(what));	\
	(void)TIFFSetErrorHandler(eHandler);	\
}

#define MAX_STRIP_BYTE_COUNT		0xffff

static ImTrace trace("ImTiff");

const char	*ImTiff::className	  = "ImTiff";
const char	*ImTiff::formatName	  = "Tiff";
ImTiff		*ImTiff::errorCurrent = NULL;

static const Im::Bitfield tiffCapabilities = IM_CAP_COLORMAP				|
											 IM_CAP_COMPRESSION				|
											 IM_CAP_MULTIPLE				|
											 IM_CAP_MUST_WRITE_LINEORDER	|
											 IM_CAP_TYPE_UNSIGNED			|
											 IM_CAP_TYPE_SIGNED				|
											 IM_CAP_TYPE_FLOAT				|
											 IM_CAP_SPACE_MONO				|
											 IM_CAP_SPACE_RGB				|
											 IM_CAP_SPACE_RGBA				|
											 IM_CAP_SPACE_CMYK				|
											 IM_CAP_BPC_8					|
											 IM_CAP_BPC_16					|
											 IM_CAP_BPC_32					|
											 IM_CAP_BPC_OTHER				|
											 IM_CAP_EXTENSION				|
											 IM_CAP_CAN_READ_PLANAR;

#ifdef IM_DSO
extern "C" {
UTDSOEXPORT void imTiffMaker( )
{
	static ImTiffMaker *maker = new ImTiffMaker( ImExemplar() );
}	//lint !e529
}
#endif

/******************************** ImTiff Options ******************************/

#include "ImOpt.h"

#define COMPRESSION_STR_LZW			"lzw"
#define COMPRESSION_STR_JPEG		"jpeg"
#define COMPRESSION_STR_PACKBITS	"packbits"
#ifdef PIXARLOG_SUPPORT
#define COMPRESSION_STR_PIXARLOG	"pixar_log"
#endif
#define COMPRESSION_STR_ZIP			"zip"

#ifdef PIXARLOG_SUPPORT
#define PIXAR_TYPE_STR_DEFAULT		"default"
#define	PIXAR_TYPE_STR_8BIT			"8_bit"
#define	PIXAR_TYPE_STR_16BIT		"16_bit"
#define	PIXAR_TYPE_STR_FLOAT		"float"
#endif

static const ImOptName compressionMenu[] =
{
	ImOptName("JPEG",		COMPRESSION_STR_JPEG),
	ImOptName("LZW",		COMPRESSION_STR_LZW),
	ImOptName("Packbits",	COMPRESSION_STR_PACKBITS),
#ifdef PIXARLOG_SUPPORT
	ImOptName("Pixar Log",	COMPRESSION_STR_PIXARLOG),
#endif
	ImOptName("Zip",		COMPRESSION_STR_ZIP),
	ImOptName(0, 0)
};

#ifdef PIXARLOG_SUPPORT
static const ImOptName pixarMenu[] =
{
	ImOptName("Default",	PIXAR_TYPE_STR_DEFAULT),
	ImOptName("8 bit",		PIXAR_TYPE_STR_8BIT),
	ImOptName("16 bit",		PIXAR_TYPE_STR_16BIT),
	ImOptName("Float",		PIXAR_TYPE_STR_FLOAT),
	ImOptName(0, 0)
};
#endif

static ImOptName names[] =
{
	ImOptName("Compression",			"compression"),
	ImOptName("Prediction encoding",	"prediction_encoding"),
	ImOptName("Jpeg quality",			"jpeg_quality"),
	ImOptName("Zip Level",				"zip_level"),
#ifdef PIXARLOG_SUPPORT
	ImOptName("Pixar Level",			"pixar_level"),
	ImOptName("Pixar Type",				"pixar_type")
#endif
};

static ImOpt theOptions[] =
{
	ImOpt(names[0],		COMPRESSION_STR_LZW, compressionMenu,	ImOptOut),
	ImOpt(names[1],		1,										ImOptOut),
	ImOpt(names[2],		1, 90, 1, 100,					ImOptOut | ImOptHardMM),
	ImOpt(names[3],		1,  5, 1,   9,					ImOptOut | ImOptHardMM),
#ifdef PIXARLOG_SUPPORT
	ImOpt(names[4],		1,  5, 1,   9,					ImOptOut | ImOptHardMM),
	ImOpt(names[5],		PIXAR_TYPE_STR_DEFAULT, pixarMenu,		ImOptIn),
#endif
	ImOpt( )
};	//lint !e1544

#define OPT_COMPRESSION			0
#define OPT_PREDICTION_ENCODING	1
#define OPT_JPEG_QUALITY		2
#define OPT_ZIP_LEVEL			3
#ifdef PIXARLOG_SUPPORT
#define OPT_PIXAR_LEVEL			4
#define OPT_PIXAR_TYPE			5
#endif
ImTiff::ImTiff( const char				*fname,
				FILE					*file,
				const Uint32			 perm ) : ImFmt( fname, file, perm )
{
	//
	// Initialize the cache, we must do this before anything else because
	// if we have a failure we must make sure the cached is NULL
	//
	cachedStrip	= 0;
	cachedSize	= 0;
	cached		= NULL;
	cacheFlushed= ImFalse;
	tiff		= NULL;

	if ( error.Occurred() )
		return;

	E_HANDLER_BEGIN;

	//
	// Since libtiff uses an fd, and we don't, we need to dup the 'fp' and
	// close the old file.  This prevents double closes or other problems.
	//
#if defined(WIN32) || defined(LINUX)
	if ( ISNULL( tiff = TIFFOpen(name, "r") ) )
	{
		E_HANDLER_END_ERROR("open");
		return;
	}
#else

	int fd;

	if ( (fd = dup(fileno(fp))) == -1 ) {
		error.V(E_NONE, E_DUP("file"));
		E_HANDLER_END;
		return;
	}

	if ( ISNULL( tiff = TIFFFdOpen( fd, name, "r" ) ) ) {
		E_HANDLER_END_ERROR("open");
		return;
	}

#endif
	//
	// close fp because we no longer need it
	//
	fclose(fp);		// TBA SYSTEM CALL
	fp = NULL;

	// allocate vars to be given to ImSpecs class
	ImArray<Uint>	 thebounds( 2 );
	ArrayChannel	 ac;
	uint16			 u16;
	Uint			 theattribs = 0;
	Uint			 numChannels;
	ImBooln			 colorMaps = ImFalse;
	ColorSpace		 colorSpace;
	DataType		 channelType;
#ifdef PIXARLOG_SUPPORT
	const char		*str;
#endif

	if ( ! TIFFGetFieldDefaulted(tiff, TIFFTAG_IMAGEWIDTH, &thebounds[0]) ) {
		error.V( E_NONE, E_TIFF_MISSING("imagewidth") );
		E_HANDLER_END;
		return;
	}

	if ( ! TIFFGetFieldDefaulted(tiff, TIFFTAG_IMAGELENGTH, &thebounds[1]) ) {
		error.V( E_NONE, E_TIFF_MISSING("imagelength") );
		E_HANDLER_END;
		return;
	}

	if ( ! TIFFGetFieldDefaulted(tiff, TIFFTAG_SAMPLESPERPIXEL, &u16) ){
		error.V( E_NONE, E_TIFF_MISSING("samplesperpixel") );
		E_HANDLER_END;
		return;
	}
	else
		numChannels = u16;

	ImArray<Uint>	channelBits( numChannels );
	ImArray<Uint>	channelSigBits( numChannels );

	if ( ! TIFFGetFieldDefaulted(tiff, TIFFTAG_BITSPERSAMPLE, &u16) ) {
		error.V( E_NONE, E_TIFF_MISSING("bitspersample") );
		E_HANDLER_END;
		return;
	} else {
		channelBits = (Uint)u16;
		channelSigBits = (Uint)u16;
	}

	if ( ! TIFFGetFieldDefaulted(tiff, TIFFTAG_PHOTOMETRIC, &u16) ) {
		error.V( E_NONE, E_TIFF_MISSING("photometric") );
		E_HANDLER_END;
		return;
	} else {
		if ( u16 == PHOTOMETRIC_RGB ) {
			if ( numChannels == 3 )
				colorSpace = ImRGB;
			else
				colorSpace = ImRGBA;
		} else if ( u16 == PHOTOMETRIC_PALETTE ) {
			colorSpace = ImRGB;
			colorMaps = ImTrue;
		} else if ( u16 == PHOTOMETRIC_MINISBLACK )
			colorSpace = ImMono;
		else if ( u16 == PHOTOMETRIC_MINISWHITE )
			colorSpace = ImInverse;
		else if ( u16 == PHOTOMETRIC_SEPARATED )
			colorSpace = ImCMYK;
		else {
			error.V( E_NONE, E_NOT_SUPPORTED( "Photometric Interpretation", u16,
											  "Tiff" ) );
			return;
		}
	}

	if ( ! TIFFGetFieldDefaulted(tiff, TIFFTAG_PLANARCONFIG, &u16) ) {
		error.V( E_NONE, E_TIFF_MISSING("planarconfig") );
		E_HANDLER_END;
		return;
	} else {
		if ( u16 == PLANARCONFIG_CONTIG )
			theattribs = IM_INTERLEAVED;
		else if ( u16 == PLANARCONFIG_SEPARATE )
			theattribs = IM_PLANAR;
	}

	if ( ! TIFFGetFieldDefaulted(tiff, TIFFTAG_COMPRESSION, &u16) ) {
		error.V( E_NONE, E_TIFF_MISSING("compression") );
		E_HANDLER_END;
		return;
	} else {
		if ( u16 != COMPRESSION_NONE )
			theattribs |= IM_COMPRESSED;
	}

	if ( TIFFIsTiled(tiff) ) {
		rowsPerStrip = 0;
		stripsInImage = 0;
		if ( ! TIFFGetField(tiff, TIFFTAG_TILEWIDTH, &tileWidth) ) {
			error.V( E_NONE, E_TIFF_MISSING("tilewidth") );
			E_HANDLER_END;
			return;
		}

		if ( ! TIFFGetField(tiff, TIFFTAG_TILELENGTH, &tileLength) ) {
			error.V( E_NONE, E_TIFF_MISSING("tilelength") );
			E_HANDLER_END;
			return;
		}
	} else {
		tileWidth  = 0;
		tileLength = 0;

		if (!TIFFGetFieldDefaulted(tiff, TIFFTAG_ROWSPERSTRIP, &rowsPerStrip)) {
			error.V( E_NONE, E_TIFF_MISSING("rowsperstrip") );
			E_HANDLER_END;
			return;
		}

		stripsInImage = TIFFNumberOfStrips(tiff);
	}

	if (!TIFFGetFieldDefaulted(tiff, TIFFTAG_SAMPLEFORMAT, &u16)) {
		channelType = Unsigned;
	} else {
		switch ( u16 ) {
		case SAMPLEFORMAT_UINT:
			channelType = Unsigned;
			break;

		case SAMPLEFORMAT_INT:
			channelType = Signed;
			break;

		case SAMPLEFORMAT_IEEEFP:
			channelType = Float;
			break;

		default:
			error.V( E_NONE, E_NOT_SUPPORTED( "sampleformat", u16, "Tiff" ) );
			E_HANDLER_END;
			return;
		}
	}

#ifdef PIXARLOG_SUPPORT
	if ( ! TIFFGetFieldDefaulted(tiff, TIFFTAG_COMPRESSION, &u16) ) {
		error.V( E_NONE, E_TIFF_MISSING("compression") );
		E_HANDLER_END;
		return;
	} else {
		switch ( u16 ) {
		case COMPRESSION_PIXARLOG:
			if ( !(str = ImFileMaker::getStrOpt(&theOptions[OPT_PIXAR_TYPE])) ||
				 !strcmp(PIXAR_TYPE_STR_DEFAULT, str) ) {
				break;		// don't set the type
			} else if ( !strcmp(PIXAR_TYPE_STR_8BIT, str) ) {
				u16 = PIXARLOGDATAFMT_8BIT;
				channelBits	   = 8U;
				channelSigBits = 8U;
				channelType	   = Unsigned;
			} else if ( !strcmp(PIXAR_TYPE_STR_16BIT, str) ) {
				u16 = PIXARLOGDATAFMT_16BIT;
				channelBits	   = 16U;
				channelSigBits = 16U;
				channelType	   = Unsigned;
			} else if ( !strcmp(PIXAR_TYPE_STR_FLOAT, str) ) {
				u16			   = PIXARLOGDATAFMT_FLOAT;
				channelBits	   = 32U;
				channelSigBits = 32U;
				channelType	   = Float;
			}
			if ( ! TIFFSetField(tiff, TIFFTAG_PIXARLOGDATAFMT, &u16) ) {
				error.V( E_NONE, E_TIFF_MISSING("compression") );
				E_HANDLER_END;
				return;
			}
			break;

		  default:
			break;
		}
	}
#endif

	//
	// Read in the colormaps LAST.  It makes it easier to delete them
	// if we have a failure.
	//
	if ( colorMaps )
	{
		uint16	*map[3];

		Uint	 mapLength = 1 << channelSigBits[0];

		// allocate space to read in color maps

		map[0] = new uint16[mapLength];
		map[1] = new uint16[mapLength];
		map[2] = new uint16[mapLength];

		if ( ISNULL( map[0] ) || ISNULL( map[1] ) || ISNULL( map[2] ) ) {
			SAFE_ARRDELETE( map[0] );
			SAFE_ARRDELETE( map[1] );
			SAFE_ARRDELETE( map[2] );
			error.V( E_NONE, E_MALLOC_FIELD( "colormap", "Tiff") );
			return;
		}

		if ( ! TIFFGetFieldDefaulted(tiff, TIFFTAG_COLORMAP, map ) ) {
			SAFE_ARRDELETE( map[0] );
			SAFE_ARRDELETE( map[1] );
			SAFE_ARRDELETE( map[2] );
			error.V( E_NONE, E_TIFF_MISSING("compression") );
			E_HANDLER_END;
			return;
		}

		// allocate channel map objects
		ChannelMap *cmap[3];
		cmap[0] = new ChannelMap( mapLength, 16 );
		cmap[1] = new ChannelMap( mapLength, 16 );
		cmap[2] = new ChannelMap( mapLength, 16 );

		if ( ISNULL( cmap[0] ) || ISNULL( cmap[1] ) || ISNULL( cmap[2] ) ) {
			SAFE_ARRDELETE( map[0] );
			SAFE_ARRDELETE( map[1] );
			SAFE_ARRDELETE( map[2] );
			SAFE_DELETE( cmap[0] );
			SAFE_DELETE( cmap[1] );
			SAFE_DELETE( cmap[2] );
			error.V( E_NONE, E_MALLOC_FIELD( "colormap", "Tiff") );
			return;
		}

		// read raw colormaps into channel map objects
		for ( Uint j = 0; j < mapLength; j++ ) {
			(*cmap[0])[j] = map[0][j];
			(*cmap[1])[j] = map[1][j];
			(*cmap[2])[j] = map[2][j];
		}

		SAFE_ARRDELETE( map[0] );
		SAFE_ARRDELETE( map[1] );
		SAFE_ARRDELETE( map[2] );

		ac.size( 3 );
		ac[0] =	cmap[0];
		ac[1] =	cmap[1];
		ac[2] =	cmap[2];
	}

	// set Im descriptor
	ImSpecs *d  = new ImSpecs( 2, thebounds, channelType, numChannels,
							   channelBits, channelSigBits, colorSpace,
							   ac, theattribs );
	(void)add( d, After );	//lint !e1506

	E_HANDLER_END;
}

ImTiff::ImTiff(
	const char			*fname,
	FILE				*file,
	const Uint32		 perm,
	const Uint			 numDims,		// number of dimensions
	const ImArray<Uint>	&dimBounds,		// bounds in each dimension
	const Uint			 numChannels,	// number of channels
	const ImArray<Uint>	&bitsPerChannel,// num of bits per channel
	const ImArray<Uint>	&significantBits,// num of significant bits per channel
	const DataType		 channelType,	// data type
	const ColorSpace	 colorSpace,	// Color space
	const ArrayChannel	&channelMaps,	// channel lookup tables
	const Uint			 storageType	// compressed?
	) : ImFmt( fname, file, perm, numDims, dimBounds, numChannels,
				bitsPerChannel, significantBits, channelType, colorSpace,
				channelMaps, storageType )
{
	uint16			 u16;	// handy value
	const char		*str;	// option value

	//
	// Initialize the cache, we must do this before anything else because
	// if we have a failure we must make sure the cached is NULL
	//
	cachedStrip	= 0;
	cachedSize	= 0;
	cached		= NULL;
	cacheFlushed= ImFalse;
	tiff		= NULL;

	if ( error.Occurred() )
		return;

	E_HANDLER_BEGIN;

	//
	// Since libtiff uses an fd, and we don't, we need to dup the 'fp' and
	// close the old file.  This prevents double closes or other problems.
	//
#if defined(WIN32) || defined(LINUX)
	//
	// close fp because we don't need it in WIN32
	//
	fclose(fp);		// TBA SYSTEM CALL
	fp = NULL;

	if ( ISNULL( tiff = TIFFOpen(name, "w") ) )
	{
		E_HANDLER_END_ERROR("open");
		return;
	}
#else
	int fd;

	if ( (fd = dup(fileno(fp))) == -1 ) {
		error.V(E_NONE, E_DUP("file"));
		E_HANDLER_END;
		return;
	}

	if ( ISNULL( tiff = TIFFFdOpen( fd, name, "w" ) ) ) {
		E_HANDLER_END_ERROR("open");
		return;
	}

	//
	// close fp because we no longer need it
	//
	fclose(fp);			// TBA SYSTEM CALL
	fp = NULL;

#endif

	fileCapabilities = (Int32)storageType;

	if ( ! TIFFSetField(tiff, TIFFTAG_IMAGEWIDTH, dimBounds[0]) ) {
		error.V( E_NONE, E_TIFF_SET("imagewidth") );
		E_HANDLER_END;
		return;
	}

	if ( ! TIFFSetField(tiff, TIFFTAG_IMAGELENGTH, dimBounds[1]) ) {
		error.V( E_NONE, E_TIFF_SET("imagelength") );
		E_HANDLER_END;
		return;
	}

	if ( ! TIFFSetField(tiff, TIFFTAG_SAMPLESPERPIXEL, (uint16)numChannels) ){
		error.V( E_NONE, E_TIFF_SET("samplesperpixel") );
		E_HANDLER_END;
		return;
	}

	if (!TIFFSetField(tiff, TIFFTAG_BITSPERSAMPLE, (uint16)bitsPerChannel[0])) {
		error.V( E_NONE, E_TIFF_SET("bitspersample") );
		E_HANDLER_END;
		return;
	}

	if ( ! validColorSpace( colorSpace, numChannels ) )
		return;

	switch ( colorSpace ) {
	case ImRGB:
		if ( numChannels == 1 )
			u16 = PHOTOMETRIC_PALETTE;
		else
			u16 = PHOTOMETRIC_RGB;
		break;

	case ImRGBA:
		u16 = PHOTOMETRIC_RGB;
		break;

	case ImCMYK:
		u16 = PHOTOMETRIC_SEPARATED;
		break;

	default:
		u16 = PHOTOMETRIC_MINISBLACK;
		break;
	}
	if ( ! TIFFSetField(tiff, TIFFTAG_PHOTOMETRIC, u16) ) {
		error.V( E_NONE, E_TIFF_SET("photometric") );
		E_HANDLER_END;
		return;
	}

	if (!TIFFSetField(tiff, TIFFTAG_PLANARCONFIG, (uint16)PLANARCONFIG_CONTIG)){
		error.V( E_NONE, E_TIFF_SET("planarconfig") );
		E_HANDLER_END;
		return;
	}

	uint16 compression;
	if ( storageType & IM_COMPRESSED ) {
		if (ISNULL(str=ImFileMaker::getStrOpt(&theOptions[OPT_COMPRESSION]))) {
			compression = COMPRESSION_LZW;
		} else if ( ! strcmp(str, COMPRESSION_STR_LZW ) ) {
			compression = COMPRESSION_LZW;
		} else if ( ! strcmp(str, COMPRESSION_STR_JPEG ) ) {
			compression = COMPRESSION_JPEG;
		} else if ( ! strcmp(str, COMPRESSION_STR_PACKBITS ) ) {
			compression = COMPRESSION_PACKBITS;
#ifdef PIXARLOG_SUPPORT
		} else if ( ! strcmp(str, COMPRESSION_STR_PIXARLOG ) ) {
			compression = COMPRESSION_PIXARLOG;
#endif
		} else if ( ! strcmp(str, COMPRESSION_STR_ZIP ) ) {
			compression = COMPRESSION_DEFLATE;
		}
	} else
		compression = COMPRESSION_NONE;

	if ( ! TIFFSetField(tiff, TIFFTAG_COMPRESSION, compression) ) {
		error.V( E_NONE, E_TIFF_SET("compression") );
		E_HANDLER_END;
		return;
	}

	switch ( compression ) {
	case COMPRESSION_DEFLATE:
		if ( !TIFFSetField(tiff, TIFFTAG_ZIPQUALITY,
			 (uint16)*ImFileMaker::getIntOpt(&theOptions[OPT_ZIP_LEVEL]))) {
			error.V( E_NONE, E_TIFF_SET("zipquality") );
			E_HANDLER_END;
			return;
		}
		/* FALL THROUGH */

	case COMPRESSION_LZW:
		if ( ImFileMaker::getTglOpt(&theOptions[OPT_PREDICTION_ENCODING]) &&
			 channelType != Float ) {
			u16			= 2;		// Horizontal prediction
			prediction	= ImTrue;
		} else {
			u16			= 1;		// no prediction
			prediction	= ImFalse;
		}

		if ( ! TIFFSetField(tiff, TIFFTAG_PREDICTOR, u16) ) {
			error.V( E_NONE, E_TIFF_SET("predictor") );
			E_HANDLER_END;
			return;
		}
		break;

	case COMPRESSION_JPEG:
		if( !TIFFSetField(tiff, TIFFTAG_JPEGQUALITY,
			(uint16)*ImFileMaker::getIntOpt(&theOptions[OPT_JPEG_QUALITY])) ) {
			error.V( E_NONE, E_TIFF_SET("jpegquality") );
			E_HANDLER_END;
			return;
		}
		break;

#ifdef PIXARLOG_SUPPORT
	case COMPRESSION_PIXARLOG:
		if ( ! TIFFSetField(tiff, TIFFTAG_PIXARLOGQUALITY,
			 (uint16)*ImFileMaker::getIntOpt(&theOptions[OPT_PIXAR_LEVEL]))) {
			error.V( E_NONE, E_TIFF_SET("pixarlogquality") );
			E_HANDLER_END;
			return;
		}
		break;
#endif

	default:
		break;
	}

	Uint bitsPerPixel = 0;
	Uint j;
	for ( j = 0; j < numChannels; j++ )
		bitsPerPixel += bitsPerChannel[j];

	Uint bytesPerPixel = (Uint)ceil( (Real)bitsPerPixel / 8. );
	Real bitsPerWord   = (Real)bytesPerPixel * 8.F;
	Real wordsPerPixel = bitsPerPixel / bitsPerWord;
	Uint bytesPerLine  = (Uint)UT_fceil(dimBounds[0]* wordsPerPixel*bytesPerPixel);
	rowsPerStrip = ImMax( ( (Uint)MAX_STRIP_BYTE_COUNT / (Uint)bytesPerLine ), (Uint)1 );

	if ( compression == COMPRESSION_JPEG ) {
		rowsPerStrip &= ~0x7;	// make multiple of 8 for jpeg
		if ( ! rowsPerStrip )
			rowsPerStrip = 8;
	}

	rowsPerStrip = ImMin( rowsPerStrip, dimBounds[1] );

	if ( ! TIFFSetField(tiff, TIFFTAG_ROWSPERSTRIP, rowsPerStrip) ) {
		error.V( E_NONE, E_TIFF_SET("rowsperstrip") );
		E_HANDLER_END;
		return;
	}

	switch ( channelType ) {
	case Unsigned:
		u16 = SAMPLEFORMAT_UINT;
		break;
	case Signed:
		u16 = SAMPLEFORMAT_INT;
		break;
	case Float:
		u16 = SAMPLEFORMAT_IEEEFP;
		break;
	default:
		break;
	}

	if ( ! TIFFSetField(tiff, TIFFTAG_SAMPLEFORMAT, u16) ) {
		error.V( E_NONE, E_TIFF_SET("sampleformat") );
		E_HANDLER_END;
		return;
	}

	//
	// Read in the colormaps LAST.  It makes it easier to delete them
	// if we have a failure.
	//
	if ( channelMaps.size() != 0 ) {

		Uint	mapLength = channelMaps[0]->size();

		Ushort *map[3];
		map[0] = new Ushort[mapLength];
		map[1] = new Ushort[mapLength];
		map[2] = new Ushort[mapLength];

		if ( ISNULL( map[0] ) || ISNULL( map[1] ) || ISNULL( map[2] ) ) {
			SAFE_ARRDELETE( map[0] );
			SAFE_ARRDELETE( map[1] );
			SAFE_ARRDELETE( map[2] );
			error.V( E_NONE, E_MALLOC_FIELD( "colormap", "Tiff" ) );
			return;
		}

		// read raw colormaps into channel map objects
		ChannelMap *redCMap   = channelMaps[0];
		ChannelMap *greenCMap = channelMaps[1];
		ChannelMap *blueCMap  = channelMaps[2];
		if	( redCMap->getSigBits() != 16 ) {
			Real bitMult = 65535.F / UT_powf(2.F, (Real)redCMap->getSigBits())-1.F;
			for ( j=0; j<mapLength; j++ ) {
				map[0][j] = Ushort( (*redCMap)[j]   * bitMult );
				map[1][j] = Ushort( (*greenCMap)[j] * bitMult );
				map[2][j] = Ushort( (*blueCMap)[j]  * bitMult );
			}
		}
		else {
			for ( j=0; j<mapLength; j++ ) {
				map[0][j] = (Ushort)(*redCMap)[j];
				map[1][j] = (Ushort)(*greenCMap)[j];
				map[2][j] = (Ushort)(*blueCMap)[j];
			}
		}

		if ( ! TIFFSetField(tiff, TIFFTAG_COLORMAP, map ) ) {
			SAFE_ARRDELETE( map[0] );
			SAFE_ARRDELETE( map[1] );
			SAFE_ARRDELETE( map[2] );
			error.V( E_NONE, E_TIFF_SET("compression") );
			E_HANDLER_END;
			return;
		}

		SAFE_ARRDELETE( map[0] );
		SAFE_ARRDELETE( map[1] );
		SAFE_ARRDELETE( map[2] );
	}

	stripsInImage = TIFFNumberOfStrips(tiff);

	E_HANDLER_END;
}

ImTiff::~ImTiff( )
{
	SAFE_ARRDELETE( cached );
	if ( tiff )
		TIFFClose(tiff);
}

ImBooln
ImTiff::cleanUp()
{
	E_HANDLER_BEGIN;

	if ( ourWritingFile )
		(void)flushCachedStrip( );	// make sure we are finished

	if ( NOTNULL( tiff ) ) {
		TIFFClose(tiff);
		tiff = NULL;
	}

	E_HANDLER_END;

	return ImTrue;
}

ImBooln
ImTiff::read( const SubArea &area, void *buffer )
{
	if ( TRACING(0) ) {									//lint !e506 !e774
		ImTrace::dprintf( "area\n" );
		area.print( ImTrace::getTraceLogFile() );
		ImTrace::dprintf( "buffer %X\n", buffer );
	}

	E_HANDLER_BEGIN;

	ImBooln	 status;
	ImBooln  all = allChannels(area);
	ImBooln  stretch = stretchBits(area);

	if ( TIFFIsTiled(tiff) || isPlanar() || !all || stretch )
		status = readScanlines( area, buffer, stretch, all );
	else
		status = readContigStripsAll( area, buffer );

	E_HANDLER_END;

	return status;
}

//
// You may notice that the code for readContigStrips is more complicated
// than the other cases.  This is because we optimize for this case.  It
// is the most common, and we want the best performance here.
//
ImBooln
ImTiff::readContigStripsAll( const SubArea &area, void *buffer )
{
	if ( area.bounds[0] == bounds()[0] ) {
		//
		// This covers the case where we don't have to do any clipping or
		// other channel re-ordering.  We do this for speed.
		//
		Uint		 line	= area.origin[1];
		Uint		 bRows	= area.bounds[1];
		Uint		 bpl	= area.bytesPerPixel() * area.bounds[0];
		tstrip_t	 strip	= line / rowsPerStrip;
		Uint		 gap	= line - (strip * rowsPerStrip);
		tsize_t		 size	= TIFFStripSize(tiff);
		Uchar		*bp		= (Uchar *)buffer;

		//
		// check for partial starting strip
		//
		if ( gap ) {
			Uchar	*cache;
			Uint	 rows = rowsInStrip(strip);

			if ( ISNULL( cache = getCachedStripRead(strip, (Uint)size)) )
				return ImFalse;

			cache += bpl * gap;				// move to the right cached row
			rows  -= gap;					// doing fewer rows

			if ( rows > bRows )				// don't do more rows than
				rows = bRows;				// we have reads

			bp = ImCopy::fullLines(cache, bp, bpl, rows);	// copy the data

			bRows -= rows;					// move on down
			strip++;
		}

		//
		// read as much as possible into the user buffer directly
		//
		while ( bRows >= rowsPerStrip ) {
			TRACE("readContigStrips", 0, ("reading strip %u", strip));
			if ( TIFFReadEncodedStrip(tiff, strip, bp, size) == -1 )
				return ImFalse;
			bp	  += size;
			bRows -= rowsPerStrip;
			strip++;
		}

		//
		// check for partial final strip
		//
		if ( bRows ) {
			Uchar	*cache;

			if ( ISNULL( cache = getCachedStripRead(strip, (Uint)size) ) )
				return ImFalse;

			(void)ImCopy::fullLines(cache, bp, bpl, bRows);	// copy the data
		}
	} else {
		//
		// This covers the case where we have to clip, but no channel
		// re-ordering.
		//
		Uint	 bpp	= area.bytesPerPixel();
		Uint	 line	= area.origin[1];
		Uint	 bRows	= area.bounds[1];
		Uint	 dBpl	= bpp * area.bounds[0];
		Uint	 sBpl	= bpp * bounds()[0];
		Uint	 sOff	= bpp * area.origin[0];
		tstrip_t strip	= line / rowsPerStrip;
		Uint	 gap	= line - (strip * rowsPerStrip);
		tsize_t	 size	= TIFFStripSize(tiff);
		Uchar	*bp		= (Uchar *)buffer;

		while ( bRows > 0 ) {
			Uchar	*cache;
			Uint	 rows = rowsInStrip(strip);

			if ( ISNULL( cache = getCachedStripRead( strip, (Uint)size) ) )
				return ImFalse;

			if ( gap ) {
				cache += sBpl * gap;		// move to the right cached row
				rows  -= gap;				// doing fewer rows
				gap	   = 0;					// no more gap
			}

			if ( rows > bRows )				// don't copy more rows than the
				rows = bRows;				// buffer has available

			bp = ImCopy::partialLines(cache + sOff, bp, sBpl, dBpl, dBpl, rows);

			bRows -= rows;					// rows are copied, move on
			strip++;						// next strip
		}
	}

	return ImTrue;
}

ImBooln
ImTiff::readScanlines( const SubArea	&area,
					   void				*buffer,
					   ImBooln			 stretch,
					   ImBooln			 all )
{
	ImBooln fullWidth = ImBooln(area.bounds[0] == bounds()[0]);

	if ( !isPlanar() && all && fullWidth && !stretch ) {
		//
		// read scanlines directly into the buffer
		//
		Uint	lines	= area.bounds[1];
		Uint	start	= area.origin[1];
		Uchar	*bp		= (Uchar *)buffer;
		Uint	 bpp	= area.bytesPerPixel();
		Uint	 bpl	= bpp * area.bounds[0];

		while ( start < lines ) {
			if ( TIFFReadScanline(tiff, bp, start++) == -1 )
				return ImFalse;

			bp += bpl;
		}
	} else if ( !isPlanar() && all && !fullWidth && !stretch ) {
		//
		// copy partial scanlines out of a buffer
		//
		Uint	 start	= area.origin[1];				// starting line
		Uint	 end	= start + area.bounds[1] - 1;	// last line
		Uint	 bpp	= area.bytesPerPixel();			// bytes per pixel
		Uint	 bpl	= bpp * area.bounds[0];			// bytes per line
		Uint	 soff	= bpp * area.origin[0];			// offset into source
		Uchar	*bp		= (Uchar *)buffer;				// pointer to buffer
		Uchar	*line	= new Uchar[(Uint)TIFFScanlineSize(tiff)];	// scanline

		if ( ISNULL( line ) ) {
			error.V(E_NONE, E_MALLOC);
			return ImFalse;
		}


		while ( start < end ) {
			if ( TIFFReadScanline(tiff, line, start++) == -1 )
				return ImFalse;

			//
			// Since we are only doing 1 line at a time, we can
			// do the simpler (fullLine) copy
			//
			bp = ImCopy::fullLines(line + soff, bp, bpl, 1);
		}

		SAFE_ARRDELETE( line );
	} else if ( !stretch ) {
		//
		// copy partial scanlines out of a buffer
		//
		Uint	 start	= area.origin[1];				// starting line
		Uint	 end	= start + area.bounds[1];		// last line
		Uchar	*line	= new Uchar[(Uint)TIFFScanlineSize(tiff)];	// scanline
		ImCopy	*copier = selectCopier(area, stretch);

		if ( ISNULL( line ) ) {
			error.V(E_NONE, E_MALLOC);
			return ImFalse;
		} else if ( ISNULL( copier ) )
			return ImFalse;

		while ( start < end ) {
			if ( TIFFReadScanline(tiff, line, start++) == -1 )
				return ImFalse;

			//
			// Since we are only doing 1 line at a time, we can
			// do the simpler (fullLine) copy
			//
			buffer = copier->copy(line, buffer, 1);
		}

		SAFE_ARRDELETE( line );
	}

	return ImTrue;
}

ImBooln
ImTiff::write( const SubArea &bufArea, void *buffer, const SubArea &outArea )
{
	if ( TRACING(0) ) {								//lint !e506 !e774
		ImTrace::dprintf( "bufArea\n" );
		bufArea.print( ImTrace::getTraceLogFile() );
		ImTrace::dprintf( "buffer %X\n", buffer );
		ImTrace::dprintf( "outArea\n" );
		outArea.print( ImTrace::getTraceLogFile() );
	}

	E_HANDLER_BEGIN;

	ImBooln	 status;

	if ( isPlanar() )
		status = ImFalse;
	else
		status = writeContigStrips( bufArea, buffer, outArea );

	E_HANDLER_END;

	return status;
}

ImBooln
ImTiff::writeContigStrips( const SubArea	&bufArea,
						   void				*buffer,
						   const SubArea	&outArea )
{
	ImBooln	all		= ImBooln(allChannels(bufArea) && allChannels(outArea));
	ImBooln stretch = stretchBits(bufArea);

	if ( all && !stretch && outArea.bounds[0] == bounds()[0] ) {
		//
		// This covers the case where we don't have to do any clipping or
		// other channel re-ordering.  We do this for speed.
		//
		Uint		 line	= outArea.origin[1];
		Uint		 bRows	= outArea.bounds[1];
		Uint		 bpl	= outArea.bytesPerPixel() * outArea.bounds[0];
		tstrip_t	 strip	= line / rowsPerStrip;
		Uint		 gap	= line - (strip * rowsPerStrip);
		tsize_t		 size	= TIFFStripSize(tiff);
		Uchar		*bp		= (Uchar *)buffer;
		Uint		 rows	= rowsInStrip(strip);
		ImBooln		 flush  = ImTrue;
		Uchar		*cache;					// use for a cache pointer


		//
		// check for partial starting strip
		//
		if ( gap ) {
			size = Int32(rows * bpl);

			if ( ISNULL( cache = getCachedStripWrite( strip, (Uint)size ) ) )
				return ImFalse;

			cache += bpl * gap;				// move to the right cached row
			rows  -= gap;					// doing fewer rows

			if ( rows > bRows )	{			// don't do more rows than
				rows  = bRows;				// we have rows to write
				flush = ImFalse;			// don't flush yet.
			}

			(void)ImCopy::fullLines(bp, cache, bpl, rows);	// copy the data
			bp += bpl * rows;

			bRows -= rows;					// move on down
			strip++;
		}

		if ( flush )
		{
			(void)flushCachedStrip( );

			//
			// write as much as possible from the user buffer directly,
			// unless we are doing prediction encoding, which blasts the
			// user buffer.  In that case, use the cache.
			//
			while ( bRows >= (rows = rowsInStrip(strip)) ) {
				size = Int32(rows * bpl);
				if ( prediction ) {
					if( ISNULL( cache=getCachedStripWrite( strip, (Uint)size ) ) )
						return ImFalse;
					(void)ImCopy::fullLines(bp, cache, bpl, rows);	// copy data
				} else {
					TRACE("writeContigStrips", 0, ("writing strip %u", strip));
					if ( TIFFWriteEncodedStrip(tiff, strip, bp, size) == -1 )
						return ImFalse;
				}
				bp	  += size;
				bRows -= rows;
				strip++;
			}

			if ( prediction )
				(void)flushCachedStrip( );	// flush the cache for prediction

			//
			// check for partial final strip
			//
			if ( bRows ) {
				Uchar	*cacheLine;

				size = Int32(rows * bpl);

				if ( ISNULL( cacheLine=getCachedStripWrite( strip, (Uint)size ) ) )
					return ImFalse;

				(void)ImCopy::fullLines(bp, cacheLine, bpl, bRows);	// copy data
			}
		}
	}

	return ImTrue;
}

Im::Uchar *
ImTiff::getCachedStripRead(Uint strip, Uint size)
{
	if ( cachedStrip != strip || !cached ) {
		if ( cachedSize != size || !cached ) {
			SAFE_ARRDELETE( cached );
			cached = new Uchar[size];
		}

		if ( cached ) {
			TRACE("getCachedStrip", 0, ("reading strip %u", strip));
			if ( TIFFReadEncodedStrip(tiff, strip, cached, (Int32)size) == -1 ) {
				SAFE_ARRDELETE( cached );
				cached		= NULL;
				cachedSize	= 0;
				cachedStrip	= 0;
			} else {
				cachedSize	= size;
				cachedStrip	= strip;
			}
		} else
			error.V( E_NONE, E_MALLOC );
	}
	else
		TRACE("getCachedStrip", 0, ("using cached strip %u", strip));

	return cached;
}

Im::Uchar *
ImTiff::getCachedStripWrite(Uint strip, Uint size)
{
	if ( cachedStrip != strip && cached && !cacheFlushed ) {
		TRACE("getCachedStripWrite", 0, ("writing strip %u", strip));
		if ( TIFFWriteEncodedStrip(tiff, cachedStrip, cached, (Int32)cachedSize )
			 == -1 ) {
			SAFE_ARRDELETE( cached );
			cached		= NULL;
			cachedSize	= 0;
			cachedStrip	= 0;

			return NULL;
		}
	}

	if ( cachedSize != size || !cached ) {
		SAFE_ARRDELETE( cached );
		if ( NOTNULL( cached = new Uchar[size] ) ) {
			cachedSize = size;
		} else {
			cachedSize = 0;
			error.V( E_NONE, E_MALLOC );
		}
	}

	cacheFlushed = ImFalse;
	cachedStrip	 = strip;

	return cached;
}

ImBooln
ImTiff::flushCachedStrip( )
{
	if ( cached && !cacheFlushed ) {
		TRACE("flushCachedStrip", 0, ("writing strip %u", cachedStrip));
		if ( TIFFWriteEncodedStrip(tiff, cachedStrip, cached, (Int32)cachedSize )
			 == -1 ) {
			SAFE_ARRDELETE( cached );
			cached		= NULL;
			cachedSize	= 0;
			cachedStrip	= 0;
			return ImFalse;
		}

		cacheFlushed = ImTrue;
	}

	return ImTrue;
}

void
ImTiff::errorHandler(const char *m, const char *f, va_list ap)
{
	char	buffer[1024], *bp = buffer;

	bp +=  sprintf(bp, "%s:", m);
	bp += vsprintf(bp, f, ap);

	errorCurrent->error.V(E_NONE, E_TIFF_ERROR(buffer));
}

const char	*ImTiff::getClassName() const { return className; }
const char	*ImTiff::getFormatName() const { return formatName; }

/*************************** ImTiffMaker *******************************/

ImTiffMaker::ImTiffMaker( ImExemplar e ) : ImFileMaker(   e,
														  ImTiff::className,
														  ImTiff::formatName,
														  "tiff",
														  tiffCapabilities,
														  10000,
														  theOptions )
{
	if ( UTlittleEndian() )
		byteSwap = ImTrue;
	else
		byteSwap = ImFalse;
}

ImTiffMaker::~ImTiffMaker()
{
}

Im *
ImTiffMaker::openIt( const char		*name,
					 FILE			*file,
					 Uint32			 perm,
					 ImBooln		 noSearch )
{
	TIFFHeader	 header;

	if ( noSearch || getHeader( file, header ) ) {
		if ( NOTNULL( file ) )
			rewind( file );

		return new ImTiff( name, file, perm );
	}

	return NULL;
}

Im *
ImTiffMaker::openIt( const char					*name,
					 FILE						*file,
				     Uint32						 permission,
				     Im::Uint					 numDims,
				     const ImArray<Im::Uint>	&bounds,
				     Im::Uint					 numChannels,
				     const ImArray<Im::Uint>	&bits,
				     const ImArray<Im::Uint>	&sigBits,
				     Im::DataType				 channelType,
				     Im::ColorSpace				 space,
				     const Im::ArrayChannel		&maps,
				     Im::Uint					 attributes )
{
	return new ImTiff( name, file, permission, numDims, bounds, numChannels,
					   bits, sigBits, channelType, space, maps, attributes );
}

static void swapDirectoryEntry( TIFFDirEntry &field ) {
	field.tdir_tag		= swapShort( field.tdir_tag );
	field.tdir_type		= swapShort( field.tdir_type );
	field.tdir_count	= swapLong( field.tdir_count );
	field.tdir_offset	= swapLong( field.tdir_offset );
}

#if (1 == LITTLE_ENDIAN)
#define TYPE3(v)	((Im::Ushort)((byteSwap == ImFalse) ? (v):((v) >> 16)))
#else
#define TYPE3(v)	((Im::Ushort)((byteSwap == ImFalse) ? ((v) >> 16):(v)))
#endif
#define TYPE4(v)	((Im::Uint)(v))
#define GET34(v)	((dtype) == TIFF_SHORT ? TYPE3(v) : TYPE4(v))

ImBooln
ImTiffMaker::getFormat(const char			* /*name*/,
					   FILE					*fp,
					   const struct stat	* /*statinfo*/,
					   Im::Uint				*width,
					   Im::Uint				*height,
					   Im::Uint				*channels,
					   Im::Uint				*bits,
					   Im::DataType			*type,
					   Im::Bitfield			*caps)
{
	TIFFHeader	 header;

	if ( getHeader( fp, header ) == ImTrue ) {
		Uint32	count = 0;

		if ( width )	{ *width = 0;			 count++; }
		if ( height )	{ *height = 0;			 count++; }
		if ( channels )	{ *channels = 1;		 count++; }
		if ( bits )		{ *bits = 1;			 count++; }
		if ( type )		{ *type = Im::Unsigned;	 count++; }
		if ( caps )		*caps	= tiffCapabilities;

		if ( !count )
			return ImTrue;

		if ( byteSwap )
			header.tiff_diroff = swapLong( header.tiff_diroff );

		// read the number of directory entries
		// TBA SYSTEM CALL
		if ( fseek( fp,	(Int32)header.tiff_diroff, SEEK_SET ) != 0 )
			return ImFalse;

		Im::Ushort numDirEntries;
		// TBA SYSTEM CALL
		if ( fread( &numDirEntries, sizeof( Im::Ushort ), 1, fp ) != 1 )
			return ImFalse;

		if ( byteSwap )
			numDirEntries = swapShort( numDirEntries );

		// allocate space for directories
		TIFFDirEntry *directories;
		if ( ISNULL( directories = (TIFFDirEntry *)
				new TIFFDirEntry[numDirEntries] ) )
			return ImFalse;

		// read directory entries
		if (fseek(fp, Int32(header.tiff_diroff+sizeof(numDirEntries)),
				  SEEK_SET ) != 0)
			return ImFalse;
		if ( fread( directories, sizeof( TIFFDirEntry ), numDirEntries, fp ) !=
			 numDirEntries )
			return ImFalse;


		for ( Uint i=0; i<numDirEntries && count; i++ ) {

			if ( byteSwap )
				swapDirectoryEntry( directories[i] );

			Im::Ushort dtype  = directories[i].tdir_type;
			Im::Uint dlength  = (Im::Uint)directories[i].tdir_count;
			Im::Uint voffset  = (Im::Uint)directories[i].tdir_offset;

			// read directory entries to get image width and height info
			switch ( directories[i].tdir_tag ) {

			case TIFFTAG_IMAGEWIDTH:
				if ( width )
				{
					count--;
					*width = GET34(voffset);
				}
				break;


			case TIFFTAG_IMAGELENGTH:
				if ( height )
				{
					count--;
					*height = GET34(voffset);
				}
				break;

			case TIFFTAG_SAMPLESPERPIXEL:
				if ( channels )
				{
					count--;
					*channels = GET34(voffset);
				}
				break;

			case TIFFTAG_BITSPERSAMPLE:
				if ( bits )
				{
					count--;
					if ( !readSampleShort(fp, dlength, voffset, bits) )
						return ImFalse;
				}
				break;

			case TIFFTAG_SAMPLEFORMAT:
				if ( type )
				{
					Im::Uint	sampleFmt = 0;

					count--;
					if ( !readSampleShort(fp, dlength, voffset, &sampleFmt) )
						return ImFalse;

					switch( sampleFmt )
					{
					case SAMPLEFORMAT_UINT:
						*type = Im::Unsigned;
						break;

					case SAMPLEFORMAT_INT:
						*type = Im::Signed;
						break;

					case SAMPLEFORMAT_IEEEFP:
						*type = Im::Float;
						break;

					case SAMPLEFORMAT_VOID:			// default to Unsigned
					default:
						break;
					}
				}
				break;

			default:	// ignore other tags
				break;
			}
		}

		SAFE_ARRDELETE( directories );

		return ImTrue;
	}

	return ImFalse;
}

ImBooln
ImTiffMaker::readSampleShort( FILE		*fp,
							  Im::Uint	 dlength,
							  Im::Uint	 voffset,
							  Im::Uint	*value )
{
	if ( dlength <= 2 ) {
		Im::Ushort	v[2];
			switch (dlength) {
			case 2: v[1] = voffset & 0xffff;
			case 1:
				if( voffset > 65535 )
					v[0] = voffset >> 16;
				else
					v[0] = voffset;
				break;
			default: break;
			}
		for ( Im::Uint j=0; j<dlength; j++ )
			if ( v[j] > *value )				//lint !e644
				*value = v[j];

		return ImTrue;
	}
	// seek to where the values reside in the file
	if ( fseek( fp,	(Int32)voffset, SEEK_SET ) != 0 ){
		return ImFalse;
	}

	Im::Ushort stmp;
	for ( Im::Uint j=0; j<dlength; j++ ) {
		// read values
		if (fread(&stmp, sizeof(Im::Ushort), 1, fp) != 1) {
			return ImFalse;
		}
		if ( byteSwap )
			stmp = swapShort( stmp );
		if ( stmp > *value )
			*value = stmp;
	}

	return ImTrue;
}

Im::Uint
ImTiffMaker::estimateSize(const ImArray<Im::Uint>	&bounds,
						  const Im::Uint			 numChannels,
						  const ImArray<Im::Uint>	&bits
						  ) const
{
	Im::Uint channelTagSize;
	if ( numChannels > 1 )
		channelTagSize = numChannels * sizeof( Im::Ushort );
	else
		channelTagSize = 0;

	Im::Uint bitsPerPixel = 0;
	for ( Im::Uint i=0; i<bits.size(); i++ )  bitsPerPixel += bits[i];

	Im::Uint bytesPerPixel = (Im::Uint)UT_fceil( (Im::Real)bitsPerPixel / 8.F );

	Im::Real bitsPerWord   = (Im::Real)bytesPerPixel * 8.F;
	Im::Real wordsPerPixel = bitsPerPixel / bitsPerWord;
	Im::Uint bytesPerLine  = (Im::Uint)UT_fceil( bounds[0] * wordsPerPixel *
											  bytesPerPixel );
	Im::Uint rowsPerStrip	= ImMax(((Uint)MAX_STRIP_BYTE_COUNT / (Uint)bytesPerLine), (Uint)1);
	rowsPerStrip			= ImMin( rowsPerStrip, bounds[1] );
	Im::Uint stripsPerImage = (Im::Uint)UT_fceil((Im::Real)bounds[1]/rowsPerStrip);

	return	sizeof( TIFFHeader )	+				// header
			sizeof( Im::Ushort )	+				// num dir entries
			( 10*sizeof( TIFFDirEntry))			 +	// directory entries
			sizeof( Im::Ulong )	+					// next ifd offset
			channelTagSize		+					// bitsPerSamples
			( stripsPerImage * sizeof(Im::Ulong)) +	// strip offsets
			( stripsPerImage * sizeof(Im::Ulong)) +	// strip byte counts
			( bounds[0] * bounds[1] *				// image size
			  numChannels *							// samples per pixel
			  (Im::Uint)ceil( bits[0] / 8. ) );		// bytesPerChannel
}

ImBooln
ImTiffMaker::isExtension( const char *ext )
{
	return ImBooln(!strncasecmp(ext, "tif", 3));
}

ImBooln
ImTiffMaker::getHeader( FILE *fp, TIFFHeader &header )
{
	rewind( fp );

	// read the Tiff file header
	if ( fread( &header, sizeof ( TIFFHeader ), 1, fp ) != 1 )
		 return ImFalse;

	Im::Ushort *bo = (Im::Ushort *)&header.tiff_magic;
	byteSwap = ImFalse;
	if ( *bo == TIFF_LITTLEENDIAN )
	{
		if ( UTlittleEndian() )
			byteSwap = ImFalse;
		else
		{
			byteSwap = ImTrue;
			header.tiff_version = swapShort(header.tiff_version);
		}
	}
	else if ( *bo == TIFF_BIGENDIAN )
	{
		if ( UTlittleEndian() )
		{
			byteSwap = ImTrue;
			header.tiff_version = swapShort(header.tiff_version);
		}
		else
			byteSwap = ImFalse;
	}
	else
		return ImFalse;

	if ( header.tiff_version == TIFF_VERSION )
		return ImTrue;

	return ImFalse;
}
