//
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
//  Produced by:
//  Silicon Grail Corporation
//  710 N. Seward
//  Hollywood, CA
//  USA 90038
//  213-871-9100
//
// NAME: ImTiff.h
//
// COMMENTS:
//	
/* $Id: ImTiff.h,v 1.1.1.1 1999/10/06 21:12:45 build Exp $ */

#ifndef IMTIFF_H
#define IMTIFF_H

#include "ImFmt.h"
#include <libtiff/tiffio.h>
#include <math.h>
#include <string.h>

class ImCopy;
class ImOpt;


class ImTiff : public ImFmt {
public:
	//
	// read constructor
	//
	ImTiff(
		const char				*fname,
		FILE					*file,
		const Uint32			 perm
		);

	//
	// write constructor
	//
	ImTiff(
		const char			*name,
		FILE				*file,
		const Uint32		 perm,
		const Uint			 numDims,		// number of dimensions
		const ImArray<Uint>	&dimBounds,		// bounds in each dimen
		const Uint			 numChannels,	// number of channels
		const ImArray<Uint>	&bitsPerChannel,// num of bits per channel
		const ImArray<Uint>	&significantBits,// num of signif. channel
		const DataType		 channelType,	// data type
		const ColorSpace	 colorSpace,	// Color space
		const ArrayChannel	&channelMaps,	// channel lookup tables
		const Uint			 storageType	// compressed?
		);

	virtual				~ImTiff();

	//
	// read, write and cleanup for file i/o
	//
	virtual	ImBooln		 read(  const SubArea &area, void *buffer );
	virtual	ImBooln		 write( const SubArea &bufArea, void *buffer,
								const SubArea &outArea );
	virtual ImBooln		 cleanUp();
	virtual const char	*getClassName() const;
	virtual const char	*getFormatName() const;

	//
	// 'code' and english name
	//
	static const char	*className;
	static const char	*formatName;

private:
	//
	// ImTiff cannot be copied or assigned.
	//
				 ImTiff(const ImTiff &);		//lint !e1704
	ImTiff		&operator=(const ImTiff &);		//lint !e1704

	//
	// Contiguous strips are layed out just like Im buffers, so
	// this is the fast path to reads.
	//
	ImBooln				 readContigStripsAll(const SubArea &area, void *buffer);

	//
	// All other image data is not in Im data order, so we handle it
	// is a slower, but more simple fashion.
	//
	ImBooln				 readScanlines( const SubArea &area, void *buffer,
										ImBooln stretch, ImBooln all);

	//
	// We only handle strip writing at this time.
	//
	ImBooln				 writeContigStrips( const SubArea	&bufArea,
											void			*buf,
											const SubArea	&outArea );

	//
	// Handle strips when we are not read/writing a full strip.
	//
	Uchar				*getCachedStripRead( Uint strip, Uint size );
	Uchar				*getCachedStripWrite( Uint strip, Uint size );
	ImBooln				 flushCachedStrip( );

	//
	// compute the number of rows in the given strip
	//
	inline Uint			 rowsInStrip( Uint strip );

	//
	// convert from libtiff to ImFiles style errors.
	//
	static void			 errorHandler(const char *m, const char *f, va_list ap);

	Uint				 tileWidth;		// tile information
	Uint				 tileLength;
	Uint				 rowsPerStrip;	// strip information
	Uint				 stripsInImage;
	Uint				 cachedStrip;	// cached data information
	Uint				 cachedSize;
	ImBooln				 cacheFlushed;
	Uchar				*cached;		// strip cache
	ImBooln				 prediction;	// prediction encoding on

	TIFF				*tiff;			// libtiff handle
	static ImTiff		*errorCurrent;	// error handling handle

};	/* end class ImTiff */

Im::Uint
ImTiff::rowsInStrip(Uint strip)
{
	Uint rows;

	if ( strip != stripsInImage-1 || (rows = bounds()[1] % rowsPerStrip) == 0 )
		rows = rowsPerStrip;

	return rows;
}

// force loading of ImTiff.o during link phase
#ifndef IM_DSO
static const char *forceTiffLoad() { return ImTiff::className; }
#endif

class ImTiffMaker : public ImFileMaker {
public:
	ImTiffMaker( ImExemplar e );
	virtual				~ImTiffMaker();

private:
	virtual Im			*openIt( const char		*name,
								 FILE			*fp,
								 Uint32			 perm,
								 ImBooln		 noSearch );

	virtual Im			*openIt( const char					*name,
								 FILE						*file,
								 Uint32						 permission,
								 Im::Uint					 numDims,
								 const ImArray<Im::Uint>	&bounds,
								 Im::Uint					 numChannels,
								 const ImArray<Im::Uint>	&bits,
								 const ImArray<Im::Uint>	&sigBits,
								 Im::DataType				 dataType,
								 Im::ColorSpace				 space,
								 const Im::ArrayChannel		&maps,
								 Im::Uint					 attributes );

	virtual ImBooln		 getFormat(const char			*name,
								   FILE					*file,
								   const struct stat	*statinfo,
								   Im::Uint				*width,
								   Im::Uint				*height,
								   Im::Uint				*channels,
								   Im::Uint				*bits,
								   Im::DataType			*type,
								   Im::Bitfield			*caps);

	virtual Im::Uint	 estimateSize( const ImArray<Im::Uint>	&bounds,
									   const Im::Uint			 numChannels,
									   const ImArray<Im::Uint>	&bits
									   ) const;

	virtual ImBooln		 isExtension( const char *ext );

	ImBooln				 getHeader( FILE *fp, TIFFHeader & );
	ImBooln				 readSampleShort( FILE		*fp,
										  Im::Uint	 dlength,
										  Im::Uint	 voffset,
										  Im::Uint	*value );

	ImBooln				 byteSwap;	// image is in opposite order
};	/* end class ImTiffMaker */

#endif /* IMTIFF_H */
