#include <ImFiles/ImFile.h>
#include <ImFiles/ImOpt.h>

//
// this should be some format that handles the same types as the format to
// be tested
//
#define OUTPUT_FORMAT	"ImTiff"

static void usage(const char *arg0, unsigned full = 0);
static void dumpFormatOptions(const char *format);
static void setFormatOption(const char  *format,
							const char  *option,
							const char **argv);
static void doReadTest( const char *fmt, int test, int argc, char **argv );
static void doWriteTest( const char *fmt, int test, int argc, char **argv );
static void doInfoTest( int test, int argc, char **argv );

static unsigned readArgc[]	= {0, 0, 1, 4, 0, 0};
static unsigned writeArgc[]	= {0, 0, 1, 0, 0, 0};

char *frame = NULL;

main(int argc, char **argv)
{
	int			 c;
	const char	*format = NULL;
	int			 readTest = -1, writeTest = -1, infoTest = -1;

	errorStderrInit( argv[0] );

#ifndef IM_STATIC
	ImFileMaker::loadFormats( );
#else
	ImFilesForceLoad( );
#endif

	while ((c = getopt(argc, argv, "f:hio:r:w:F:LO")) != -1) {
		switch( c ) {
		case 'f':
			format = optarg;
			break;

		case 'h':
			usage(argv[0], 1);
			exit(0);
			break;

		case 'i':
			infoTest = 1;
			break;

		case 'o':
			setFormatOption( format, optarg, argv );
			break;

		case 'r':
			readTest = atoi(optarg);
			break;

		case 'w':
			writeTest = atoi(optarg);
			break;

		case 'F':
			frame = optarg;
			break;

		case 'L':
		{
			const char *fmtClass, *fmtName, *fmtExtension;

			while ( ImFileMaker::formatNamesSorted( &fmtClass, &fmtName,
													&fmtExtension ) )
			{
				fprintf(stderr, "%-12s %-24s %-10s\n", fmtClass, fmtName,
						fmtExtension);
			}

			exit(0);
		}

		case 'O':
			if ( !format ) {
				fprintf(stderr, "You must specify the format with the -f option for the -O option to work.\n");
				exit(1);
			}
			dumpFormatOptions(format);
			exit(0);
			break;

		default:
			break;
		}
	}

	if ( (readTest == -1 && writeTest == -1 && infoTest == -1 ) ||
		 (readTest != -1 && writeTest != -1) ||
		 (readTest != -1 && infoTest != -1) ||
		 (infoTest != -1 && writeTest != -1) ) {
		fprintf(stderr, "must specify one (and only one) read, write or info test\n");
		usage(argv[0]);
		exit(1);
	}

	if ( readTest != -1 )
		doReadTest( format, readTest, argc, argv );
	else if ( writeTest != -1 )
		doWriteTest( format, writeTest, argc, argv );
	else
		doInfoTest( infoTest, argc, argv );
}

static const char * const options[2] = {
	"usage: %s [-F frame][-f fmt][-h][-i][-o option value][-O][-r #][-w #] ...\n",
	"\n"
	"-F\tSet the current frame # to frame.\n"
	"-f\tSet the class name of the format to test.\n"
	"-h\tThis help sheet.\n"
	"-i\tPerform a get format/information test.\n"
	"-L\tDump the list of installed formats.\n"
	"-o\tSet the format option to value.\n"
	"-O\tDump a list of format options.\n"
	"-r\tPerform read test #.\n"
	"-w\tPerform write test #.\n"
};

static void
usage(const char *arg0, unsigned full)
{
	const char *cmd = strchr(arg0, '/');
	if ( cmd == NULL )
		cmd = arg0;

	fprintf(stderr, options[0], cmd);

	if ( !full )
		return;

	fprintf(stderr, options[1]);
}

static void
dumpFormatOptions(const char *format)
{
	const ImOpt	*options = ImFileMaker::getOptionsDescription(format);

	if ( options ) {
		ImOpt::Type	type;
		while ( type = options->getType( ) ) {
			unsigned flags = options->getFlags();
			if (flags & ImOptIn)
				printf("I");
			else
				printf(" ");
			if (flags & ImOptOut)
				printf("O ");
			else
				printf("  ");

			switch( type ) {
			case ImOpt::tToggle:
				printf("Toggle '%s' default=%s\n", options->getTag(),
					   options->getDefaultToggle() ? "True" : "False");
				break;

			case ImOpt::tInteger: {
				int	*defaults	= options->getDefaultInteger();
				int	 count		= options->getCountInteger();

				printf("Integer '%s' count=%d, default=(",
					   options->getTag(), count );

				for ( int i = 0; i < count - 1; i++ )
					printf( "%d, ", defaults[i] );
				printf( "%d), ", defaults[count - 1] );

				printf("%smin=%d, %smax=%d\n",
					   flags&ImOptHardMin ?"hard ":"",options->getMinInteger(),
					   flags&ImOptHardMax ?"hard ":"",options->getMaxInteger());
			}	break;

			case ImOpt::tFloat: {
				float	*defaults	= options->getDefaultFloat();
				int		 count		= options->getCountFloat();

				printf("Float '%s' count=%f, default=(",
					   options->getTag(), count);
				
				for ( int i = 0; i < count - 1; i++ )
					printf( "%f, ", defaults[i] );
				printf( "%f), ", defaults[count - 1] );

				printf("%smin=%f, %smax=%f\n",
					   flags&ImOptHardMin ? "hard ":"", options->getMinFloat(),
					   flags&ImOptHardMax ? "hard ":"", options->getMaxFloat());
			}	break;

			case ImOpt::tString: {
				const ImOptName	*menu = options->getStringMenu();
				printf("String '%s' default=%s, menu {", options->getTag(),
					   options->getDefaultString());
				if ( menu ) {
					while ( menu->myTag ) {
						printf( "%s", menu->myTag );
						menu++;
						if ( menu->myTag )
							printf( ", " );
					}
				}
				printf("}\n");
				break;
			}
			}

			options++;
		}
	}
}

static void
setFormatOption(const char *format, const char *option, const char **argv)
{
	const ImOpt	*options = ImFileMaker::getOptionsDescription(format);

	if ( options ) {
		ImOpt::Type	type;
		while ( type = options->getType( ) ) {
			if ( !strcmp(options->getTag(), option) ) {
				switch( type ) {
				case ImOpt::tToggle:
					ImFileMaker::setOption(options, *argv[optind++] == 't');
					break;

				case ImOpt::tInteger: {
					unsigned	 count = options->getCountInteger();
					int			*ints  = new int[count];

					for ( unsigned i = 0; i < count; i++ )
						ints[i] = atoi(argv[optind++]);

					ImFileMaker::setOption(options, ints);

					delete [] ints;
					break;
				}

				case ImOpt::tFloat: {
					unsigned	 count	= options->getCountInteger();
					float		*floats = new float[count];

					for ( unsigned i = 0; i < count; i++ )
						floats[i] = atof(argv[optind++]);

					ImFileMaker::setOption(options, floats);

					delete [] floats;
					break;
				}

				case ImOpt::tString:
					ImFileMaker::setOption(options, argv[optind++]);
					break;
				}
			}
			options++;
		}
	}
}

static void
doReadTest( const char *format, int test, int argc, char **argv )
{
	Im			*in, *out;
	Im::Uint	 attributes = IM_COMPRESSED | IM_INTERLEAVED;
	int			 argind = optind;

	optind += readArgc[test];
	if ( optind >= argc ) {
		errorV( E_FATAL( errorStderr ), 0, "Not enough arguments.\n" );
		exit( 1 );
	}

	if ( format )
		in = ImFileMaker::open(argv[optind++], NULL, format, IM_FILE_READ);
	else
		in = ImFileMaker::open(argv[optind++], NULL, IM_FILE_READ);
	if ( in == NULL || in->error.Occurred() ) {
		errorV( E_FATAL( errorStderr ), 0, "Can't open %s for input",
				argv[optind - 1]);
		exit(1);
	}

	if ( frame && !in->seek(frame) ) {
		errorV( E_FATAL( errorStderr ), 0, "Can't seek to frame %s in %s",
				frame, argv[optind - 1]);
		exit(1);
	}

	Im::SubArea	 inArea(*in);
	Im::SubArea  outArea(inArea);

	if ( outArea.sigBitsPerChannel[0] <= 8 )
		outArea.sigBitsPerChannel = 8;
	else if ( outArea.sigBitsPerChannel[0] <= 16 )
		outArea.sigBitsPerChannel = 16;
	else if ( outArea.sigBitsPerChannel[0] <= 32 )
		outArea.sigBitsPerChannel = 32;

	if ( test == 3 ) {
		Im::Uint x1 = (Im::Uint)atoi(argv[argind++]);
		Im::Uint y1 = (Im::Uint)atoi(argv[argind++]);
		Im::Uint x2 = (Im::Uint)atoi(argv[argind++]);
		Im::Uint y2 = (Im::Uint)atoi(argv[argind++]);
		Im::Uint t;

		if ( x1 > x2 ) {
			t = x1;
			x1 = x2;
			x2 = t;
		}

		if ( y1 > y2 ) {
			t = y1;
			y1 = y2;
			y2 = t;
		}

		inArea.origin[0] = x1;
		inArea.origin[1] = y1;

		inArea.bounds[0] = x2 - x1 + 1;
		inArea.bounds[1] = y2 - y1 + 1;

		outArea.bounds[0] = inArea.bounds[0];
		outArea.bounds[1] = inArea.bounds[1];
	}

	out = ImFileMaker::open(argv[optind++], NULL, OUTPUT_FORMAT,
							IM_FILE_WRITE | IM_FILE_CREATE,
							in->dims(), outArea.bounds, in->channels(),
							outArea.sigBitsPerChannel,
							outArea.sigBitsPerChannel,
							in->type(), in->space(), Im::ArrayChannel(),
							attributes);

	if ( out == NULL || out->error.Occurred() ) {
		errorV( E_FATAL( errorStderr ), 0, "Can't open %s for output",
				argv[optind - 1]);
		exit(1);
	}

	Im::Uchar	*buffer = new Im::Uchar[inArea.bytesPer()];

	switch( test )
	{
	//
	// read file into buffer in ABGR order
	// NOTE TO CHALICE PLUGIN DEVELOPERS:  You do NOT have to be able to
	// handle test 5, this is for internal use ONLY.
	//
	case 5: {
		for ( Im::Uint i = 0; i < inArea.whichChannels.size(); i++ )
			inArea.whichChannels[i] = inArea.whichChannels.size() - i - 1;
	}	/*FALLTHRU*/

	//
	// Read a sub area and write it.
	//
	case 3:
		/*FALLTHRU*/

	//
	// Read the file as one large RGB image.
	//
	case 1: {
		if ( ! in->read(inArea, buffer) ) {
			errorV( E_FATAL( errorStderr ), 0, "Can't read %s",
					in->getFileName());
			exit(1);
		}
		if ( ! out->write(outArea, buffer, outArea) ) {
			errorV( E_FATAL( errorStderr ), 0, "Can't write %s",
					in->getFileName());
			exit(1);
		}
		break;
	}

	//
	// read the image in N slices
	//
	case 2: {
		Im::Uint	 slices = (Im::Uint)atoi(argv[argind++]);
		Im::Uint	 lines = inArea.bounds[1] / slices;
		Im::Uchar	*bp = buffer;

		if ( lines < 1 )
			lines = 1;

		inArea.bounds[1] = lines;

		for ( Im::Uint y = 0; y < outArea.bounds[1]; y += lines ) {
			inArea.origin[1] = y;

			if ( y + lines > outArea.bounds[1] )
				inArea.bounds[1] = outArea.bounds[1] - y;

			if ( ! in->read(inArea, bp) ) {
				errorV( E_FATAL( errorStderr ), 0, "Can't read slice %d of %s",
						(int)y / (int)lines, in->getFileName());
				exit(1);
			}

			bp += inArea.bytesPer();			// move to next slice in memory
		}
		if ( ! out->write(outArea, buffer, outArea) ) {	// write the whole thing
			errorV( E_FATAL( errorStderr ), 0, "Can't write %s",
					in->getFileName());
			exit(1);
		}
		break;
	}

	default:
		break;
	}

	delete [] buffer;

	if ( ! in->cleanUp() ) {
		errorV( E_FATAL( errorStderr ), 0, "Can't cleanup %s",
				in->getFileName());
		exit(1);
	}

	if ( ! out->cleanUp() ) {
		errorV( E_FATAL( errorStderr ), 0, "Can't cleanup %s",
				out->getFileName());
		exit(1);
	}

	delete in;
	delete out;
}

static void
doWriteTest( const char *format, int test, int argc, char **argv )
{
	Im			*in, *out;
	Im::Uint	 attributes = IM_COMPRESSED | IM_INTERLEAVED;
	int			 argind = optind;

	optind += writeArgc[test];
	if ( optind >= argc ) {
		errorV( E_FATAL( errorStderr ), 0, "Not enough arguments.\n" );
		exit( 1 );
	}

	in = ImFileMaker::open(argv[optind++], NULL, IM_FILE_READ);
	if ( in == NULL || in->error.Occurred() ) {
		errorV( E_FATAL( errorStderr ), 0, "Can't open %s for input",
				argv[optind - 1]);
		exit(1);
	}

	Im::SubArea inArea(*in);
	Im::SubArea	outArea(inArea);

 	if ( outArea.sigBitsPerChannel[0] <= 8 )
 		outArea.sigBitsPerChannel = 8;
 	else if ( outArea.sigBitsPerChannel[0] <= 16 )
 		outArea.sigBitsPerChannel = 16;
 	else if ( outArea.sigBitsPerChannel[0] <= 32 )
		outArea.sigBitsPerChannel = 32;

	out = ImFileMaker::open(argv[optind++], NULL, format,
							IM_FILE_WRITE | IM_FILE_CREATE,
							in->dims(), in->bounds(), in->channels(),
							outArea.sigBitsPerChannel,
							outArea.sigBitsPerChannel,
							in->type(), in->space(), Im::ArrayChannel(),
							attributes);

	if ( out == NULL || out->error.Occurred() ) {
		errorV( E_FATAL( errorStderr ), 0, "Can't open %s for output",
				argv[optind - 1]);
		exit(1);
	}

	if ( frame && !out->seek(frame) ) {
		errorV( E_FATAL( errorStderr ), 0, "Can't seek to frame %s in %s",
				frame, argv[optind - 1]);
		exit(1);
	}

	Im::Uchar *buffer = new Im::Uchar[inArea.bytesPer()];

	switch( test )
	{
	case 1:
		if ( ! in->read(inArea, buffer) ) {
			errorV( E_FATAL( errorStderr ), 0, "Can't read %s",
					in->getFileName());
			exit(1);
		}
		if ( ! out->write(inArea, buffer, inArea) ) {
			errorV( E_FATAL( errorStderr ), 0, "Can't write %s",
					in->getFileName());
			exit(1);
		}
		break;

	//
	// write the image in N slices
	//
	case 2: {
		Im::SubArea	 outArea(inArea);
		Im::Uint	 slices = (Im::Uint)atoi(argv[argind++]);
		Im::Uint	 lines = outArea.bounds[1] / slices;
		Im::Uchar	*bp = buffer;

		if ( lines < 1 )
			lines = 1;

		outArea.bounds[1] = lines;

		if ( ! in->read(inArea, buffer) ) {
			errorV( E_FATAL( errorStderr ), 0, "Can't read %s",
					in->getFileName());
			exit(1);
		}

		for ( Im::Uint y = 0; y < inArea.bounds[1]; y += lines ) {
			outArea.origin[1] = y;

			if ( y + lines > inArea.bounds[1] )
				outArea.bounds[1] = inArea.bounds[1] - y;

			if ( ! out->write(outArea, bp, outArea) ) {
				errorV( E_FATAL( errorStderr ), 0, "Can't write slice %d of %s",
						(int)y / (int)lines, in->getFileName());
				exit(1);
			}

			bp += outArea.bytesPer();			// move to next slice in memory
		}

		break;
	}

	default:
		break;
	}

	delete [] buffer;

	if ( ! in->cleanUp() ) {
		errorV( E_FATAL( errorStderr ), 0, "Can't cleanup %s",
				in->getFileName());
		exit(1);
	}

	if ( ! out->cleanUp() ) {
		errorV( E_FATAL( errorStderr ), 0, "Can't cleanup %s",
				out->getFileName());
		exit(1);
	}

	delete in;
	delete out;
}

static void
doInfoTest( int /*test*/, int argc, char **argv )
{
	Im::Uint		 width;
	Im::Uint		 height;
	Im::Uint		 channels;
	Im::Uint		 bits;
	Im::DataType	 type;
	Im::Uint		 size;

	if ( optind >= argc ) {
		errorV( E_FATAL( errorStderr ), 0, "Not enough arguments.\n" );
		exit( 1 );
	}

	const char *file = argv[optind++];
	const char *format = ImFileMaker::format(file, ImTrue, &width, &height,
											 &channels, &bits, &type, &size);

	if ( format == NULL ) {
		errorV( E_FATAL( errorStderr ), 0, "Can't getFormat for %s", file );
		exit(1);
	}

	printf( "file='%s', format='%s', width=%u, height=%u, channels=%u,\n"
			"bits=%u, type=%s, size=%u\n", file, format, width, height,
			channels, bits, Im::typeTable[type], size);
}
