/*
 * Dither plugin
 *   This plugin dithers an image using Floyd and Steinberg
 * error diffusion dithering.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include "cpi.h"

/*
 * Simple inline function to add the error to some value,
 * while clamping it between 0 and max
 */
template <class T>
inline void AddError( T &val, const float error, const T max )
{
	if( error < 0.0F )
	{
		T e = (T)( (float)-error );
		if( val > max - e )
			val = max;
		else
			val = val + e;
	}
	else
	{
		T e = error;
		if( val < e )
			val = 0;
		else
			val = val - e;
	}
}

/*
 * Dither
 *   Actually does the real work.  It uses the Floyd-Steinberg error
 * diffusion alogarithm.  It computes a simple threshold for the current
 * pixel, then propagates the error to the next pixel, and the three pixels
 * below the current pixel in the following ratios: 7/16, 3/16, 5/16, and
 * 1/16.
 */
template <class T>
void Dither( CPI_Image *out, CPI_Image *in, T max, int levels )
{
	int x, y, c;
	T *outdata = (T *)out->data;
	T *indata = (T *)in->data;
	float error;
	T p;

	int sizeX = out->info.sizeX;
	int sizeY = out->info.sizeY;
	int channels = out->info.channels;
	int pix;

	/* Run over all pixels of the output image */
	for( y = 0; y<sizeY; y++ )
		for( x = 0; x<sizeX; x++ )
		{
			pix = (y*sizeX+x)*channels;
			c = 0;
			do
			{
				/* Threshold the pixel */
				p = indata[pix+c];
				if( levels > 1 )
					outdata[pix+c] = ((int)(p * levels / max) ) * max / levels;
				else
					outdata[pix+c] = p<max/2 ? 0 : max;

				/* Calculate the error */
				error = (float)( outdata[pix+c] - p);

				/* Propagate the error to next pixels */
				if( x < sizeX-1 )
					AddError( indata[pix+c+channels], (error*7/16), max );
				if( y < sizeY-1 )
					AddError( indata[pix+c+channels*(sizeX)], (error*5/16), max );
				if( x > 0 && y < sizeY-1 )
					AddError( indata[pix+c+channels*(sizeX-1)], (error*3/16), max );
				if( y < sizeY-1 && x < sizeX-1 )
					AddError( indata[pix+c+channels*(sizeX+1)], (error/16), max );
				c++;
			}
			while( c<channels );
		}
}

/*
 * Only a single parameter: number of levels
 */
CPIDSOEXPORT void upiCreateParameters( void )
{
	cpiAddTab( "Dither" );
	cpiAddInteger( "Levels", 2, 2, 10, P_MIN );
	return;
}


/*
 * We don't have a way to skip the Floyd-Steinberg error stuff,
 * so let's make things cook full
 */
CPIDSOEXPORT int
upiCookFull( void )
{
	return 1;
}

/*
 * The actually processing function
 * This figures out what type of pixels we have and just
 * calls the templated Tile function at the top
 */
CPIDSOEXPORT int upiProcessImage( CPI_Image *result )
{
	int levels;
	cpiGetInteger( "Levels",result->info.time, &levels);
	if( levels < 2 )
		levels = 2;

	levels--;

	switch( result->info.pelType )
	{
		case P_INT8:
			Dither( result, result, (unsigned char)UCHAR_MAX, levels );
			break;
		case P_INT16:
			Dither( result, result, (unsigned short)USHRT_MAX, levels );
			break;
		case P_FLOAT32:
			Dither( result, result, (float)1.0F, levels );
			break;
		default:
			cpiError( "Unknown pixel type" );
			return 1;
	}
	return 0;
}

