/*
 * User Plugin Interface
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <math.h>
#include "cpi.h"

/* Random number generator:
 *   Straight from the book "Numerical Recipes in C", chapter 7.
 */
#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

/* Return a random uniform deviate between 0.0 and 1.0 */
float uniformRandom( long &seed )
{
	int j;
	long k;
	static long iy = 0;
	static long iv[NTAB];
	float temp;
	if( seed <= 0 || !iy )
	{
		if( -seed < 1 ) seed = 1;
		else seed = -seed;
		for( j=NTAB+7; j>=0; j-- )
		{
			k = seed/IQ;
			seed = IA*(seed-k*IQ)-IR*k;
			if( seed < 0 ) seed += IM;
			if( j < NTAB ) iv[j] = seed;
		}
		iy=iv[0];
	}
	k = seed/IQ;
	seed = IA*(seed-k*IQ)-IR*k;
	if( seed < 0 ) seed += IM;
	j = iy/NDIV;
	iy = iv[j];
	iv[j] = seed;
	if( (temp=AM*iy) > RNMX ) return RNMX;
	return temp;
}

/* Returns a normally distributed deviate with zero mean and unit variance */
float gaussianRandom( long &seed )
{
	static int iset = 0;
	static float gset = 0.0F;
	float fac, rsq, v1, v2;
	if( iset == 0 )
	{
		rsq = 2.0F;
		while( rsq >= 1.0F )
		{
			v1 = 2.0*uniformRandom(seed) - 1.0;
			v2 = 2.0*uniformRandom(seed) - 1.0;
			rsq = v1 * v1 + v2 * v2;
		}
		fac = sqrt(-2.0*log(rsq)/rsq);
		gset = v1 * fac;
		iset = 1;
		return v2 * fac;
	}
	iset = 0;
	return gset;
}

template <class T>
void Gaussian( T *image, long seed, int pixels,
				float max, float amount, int type )
{
	float num;
	if( type != 0 )
	{
		while ( pixels-- )
		{
			num = (float)(*image) * gaussianRandom(seed) * amount;
			if ( num < 0.0F ) num = 0.0F;
			if ( num > max ) num = max;

			*image = (T)num;
			image++;
		}
	}
	else
	{
		while( pixels-- )
		{
			num = (float)(*image) + (gaussianRandom(seed)*max*amount);
			if( num < 0.0F ) num = 0.0F;
			if( num > max ) num = max;
			*image = (T)num;
			image++;
		}
	}
}

// let's provide a particular implemenation for float which doesn't clip...
void Gaussian( float *image, long seed, int pixels,
				float max, float amount, int type )
{
	if( type != 0 )
	{
		while ( pixels-- )
		{
			*image = *image * gaussianRandom(seed) * amount;
			image++;
		}
	}
	else
	{
		while( pixels-- )
		{
			*image = *image + (gaussianRandom(seed)*max*amount);
			image++;
		}
	}
}

char *menu[] = { "Add", "Multiply", NULL };

CPIDSOEXPORT void upiCreateParameters( void )
{
	cpiAddTab( "Noise" );
	cpiAddInteger( "Seed", 1, 0, 100, P_MIN | P_ANIMATE );
	cpiAddFloat( "Amount", 0.5, 0.0, 1.0, P_ANIMATE );
	cpiAddMenu( "Mix", 0, menu, 0 );
	return;
}

/*
 * We don't have a way to skip noise, so let's make things cook full
 */
CPIDSOEXPORT int
upiCookFull( void )
{
	return 1;
}
 
/*
 * Do the actual work
 */
CPIDSOEXPORT int upiProcessImage( CPI_Image *result )
{
	int s;
	long seed;
	int type;
	float amount;

	/* Get the parameters */
	cpiGetInteger( "Seed", result->info.time, &s );
	cpiGetInteger( "Mix", result->info.time, &type );
	cpiGetFloat( "Amount", result->info.time, &amount );

	seed = -s;
	int pixels = result->info.sizeX * result->info.sizeY * result->info.channels;

	/* Call the right template function for the type of image we have */
	switch( result->info.pelType )
	{
		case P_INT8:
			Gaussian( (unsigned char *)result->data, seed, pixels,
						(float)UCHAR_MAX, amount, type );
			break;
		case P_INT16:
			Gaussian( (unsigned short *)result->data, seed, pixels,
						(float)USHRT_MAX, amount, type );
			break;
		case P_FLOAT32:
			Gaussian( (float *)result->data, seed, pixels,
						(float)1.0F, amount, type );
			break;
		default:
			cpiError( "Unknown pixel type" );
			return 1;
	}
	return 0;
}

