/*
 * User Plugin Interface
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include "cpi.h"

/* Data that gets passed around to all thread when multi-processing */
template <class T>
struct MaxData
{
	T *max;
	T *min;
	T **start;
	T **end;
};

/* Multi-process function that finds the max and min of an area */
void MultiProcMaxFloat( int proc, int numProc, void *arg )
{
	/* Initialize what I have to do */
	MaxData<float> *data = (MaxData<float> *)arg;
	float *start = data->start[proc];
	float *end = data->end[proc];
	float max = 0.0F;
	float min = 999999.0F;
	
	/* Find the maximum and minimum for my pixels */
	while( start < end )
	{
		if( *start > max )
			max = *start;
		if( *start < min )
			min = *start;
		start++;
	}

	/* And we're done */
	data->max[proc] = max;
	data->min[proc] = min;
}

void MultiProcMaxShort( int proc, int numProc, void *arg )
{
	MaxData<unsigned short> *data = (MaxData<unsigned short> *)arg;
	unsigned short *start = data->start[proc];
	unsigned short *end = data->end[proc];
	unsigned short max = 0;
	unsigned short min = USHRT_MAX;
	
	while( start < end )
	{
		if( *start > max )
			max = *start;
		if( *start < min )
			min = *start;
		start++;
	}

	data->max[proc] = max;
	data->min[proc] = min;
}

void MultiProcMaxChar( int proc, int numProc, void *arg )
{
	MaxData<unsigned char> *data = (MaxData<unsigned char> *)arg;
	unsigned char *start = data->start[proc];
	unsigned char *end = data->end[proc];
	unsigned char max = 0;
	unsigned char min = UCHAR_MAX;
	
	while( start < end )
	{
		if( *start > max )
			max = *start;
		if( *start < min )
			min = *start;
		start++;
	}

	data->max[proc] = max;
	data->min[proc] = min;
}

/* Multi-process function which scales all values to be between min and max */
void MultiProcScaleFloat( int proc, int numProc, void *arg )
{
	MaxData<float> *data = (MaxData<float> *)arg;

	float *s = data->start[proc];
	float *e = data->end[proc];
	float max = data->max[proc];
	float min = data->min[proc];

	if( max <= min )
	{
		while( s < e )
		{
			*s = (float)0;
			s++;
		}
	}
	else
	{
		max -= min;
		while( s < e )
		{
			*s = (float)( ((float)*s - (float)min)/(float)max );
			s++;
		}
	}
}

void MultiProcScaleShort( int proc, int numProc, void *arg )
{
	MaxData<unsigned short> *data = (MaxData<unsigned short> *)arg;

	unsigned short *s = data->start[proc];
	unsigned short *e = data->end[proc];
	unsigned short max = data->max[proc];
	unsigned short min = data->min[proc];

	if( max <= min )
	{
		while( s < e )
		{
			*s = (unsigned short)0;
			s++;
		}
	}
	else
	{
		max -= min;
		while( s < e )
		{
			*s = (unsigned short)( ((float)*s - (float)min)*USHRT_MAX/(float)max );
			s++;
		}
	}
}

void MultiProcScaleChar( int proc, int numProc, void *arg )
{
	MaxData<unsigned char> *data = (MaxData<unsigned char> *)arg;

	unsigned char *s = data->start[proc];
	unsigned char *e = data->end[proc];
	unsigned char max = data->max[proc];
	unsigned char min = data->min[proc];

	if( max <= min )
	{
		while( s < e )
		{
			*s = (unsigned char)0;
			s++;
		}
	}
	else
	{
		max -= min;
		while( s < e )
		{
			*s = (unsigned char)( ((float)*s - (float)min)*UCHAR_MAX/(float)max );
			s++;
		}
	}
}

/*
 * Perform the normalize operation.
 *   This calls the above multi-processed functions
 */
template <class T>
int Normalize( CPI_Image *in, CPI_Image *out, T MAX )
{
	int procs = cpiNumProcessors();

	// Create a data structure to multi-proc
	MaxData<T> data;
	data.max = new T[procs];
	data.min = new T[procs];
	data.start = new T *[procs];
	data.end = new T *[procs];

	// Split up the image into N pieces
	int len = in->info.sizeX * in->info.sizeY * in->info.channels;
	int part = len/procs;
	int i;
	for( i=0; i<procs; i++ )
	{
		data.start[i] = ((T *)in->data) + part * i;
		data.end[i] = ((T *)in->data) + part * (i+1);
	}
	i--;
	data.end[i] = ((T *)in->data) + len;

	switch( sizeof(T) )
	{
		case 1:
			cpiMultiProcess( MultiProcMaxChar, (void *)(&data) );
			break;

		case 2:
			cpiMultiProcess( MultiProcMaxShort, (void *)(&data) );
			break;

		case 4:
			cpiMultiProcess( MultiProcMaxFloat, (void *)(&data) );
			break;

		default:
			/* ?? */
			break;
	}

	// Find the max and min from all the threads
	T max = 0;
	T min = MAX;
	for( i=0; i<procs; i++ )
	{
		if( data.min[i] < min )
			min = data.min[i];
		if( data.max[i] > max )
			max = data.max[i];
	}

	for( i=0; i<procs; i++ )
	{
		data.min[i] = min;
		data.max[i] = max;
	}

	// Now scale all the values
	switch( sizeof(T) )
	{
		case 1:
			cpiMultiProcess( MultiProcScaleChar, (void *)(&data) );
			break;

		case 2:
			cpiMultiProcess( MultiProcScaleShort, (void *)(&data) );
			break;

		case 4:
			cpiMultiProcess( MultiProcScaleFloat, (void *)(&data) );
			break;

		default:
			/* ?? */
			break;
	}

	return 0;
}

/* We can't slice or do ROI */
CPIDSOEXPORT int upiCookFull( void )
{
	return 1;
}

/* Do the image processing */
CPIDSOEXPORT int upiProcessImage( CPI_Image *result )
{
	switch( result->info.pelType )
	{
		case P_INT8:
		{
			Normalize( result, result, (unsigned char)UCHAR_MAX );
			break;
		}
		case P_INT16:
		{
			Normalize( result, result, (unsigned short)USHRT_MAX );
			break;
		}
		case P_FLOAT32:
		{
			Normalize( result, result, (float)1.0F );
			break;
		}
		default:
			cpiError( "Unknown pixel type" );
			return 1;
	}
	return 0;
}
