/*
 * Tile plugin
 *   This plugin will repeat the image in both X and Y dimensions
 * It resizes the output image.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include "cpi.h"

/*
 * This does the actual tiling work.
 * It is templated to handle 8 bit, 16 bit and float data.
 * It's currently not optimized, but the compiler should
 * be good enough to make this run pretty fast.
 */
template <class T>
void Tile( CPI_Image *out, CPI_Image *in, T max )
{
	int x, y, c;
	T *img = (T *)out->data;
	T *inp = (T *)in->data;

	int sizeX = out->info.sizeX;
	int sizeY = out->info.sizeY;
	int channels = out->info.channels;

	int inX = in->info.fullX;
	int inY = in->info.fullY;
	int rx, ry;
	int inpix, outpix;

	/* Run over all pixels of the output image */
	for( y = 0; y<sizeY; y++ )
	{
		for( x = 0; x<sizeX; x++ )
		{
			/* Map the output pixel to an input pixel */
			rx = (x+out->info.offsetX)%inX;
			ry = (y+out->info.offsetY)%inY;
			inpix = (ry*inX+rx)*channels;
			outpix = (y*sizeX+x)*channels;
			/* Copy all channels from the input pixel to the output one. */
			c = 0;
			do
			{
				img[outpix+c] = inp[inpix+c];
				c++;
			}
			while( c<channels );
		}
	}
}

/* 
 * The parameters
 *  Tile X - number of times to repeat the image in X direction
 *  Tile Y - number of times to repeat the image in Y direction
 */
CPIDSOEXPORT void upiCreateParameters( void )
{
	cpiAddTab( "Tile" );
	cpiAddFloat( "Tile X", 2, 0, 10, P_MIN );
	cpiAddFloat( "Tile Y", 2, 0, 10, P_MIN );
	return;
}

/*
 * When our parameters (Tile X and Tile Y) change, we 
 * need to tell chalice our output info has changed.
 */
CPIDSOEXPORT void upiParameterChanged( char *name )
{
	cpiInvalidateOutput();
}

/*
 * We never use the input as our result image
 */
CPIDSOEXPORT int upiResultInput( CPI_ImageContext *result )
{
	return RESULT_SEPARATE;
}

/*
 * Calculate our output size from the input size
 * and parameters.
 */
CPIDSOEXPORT int upiOutputContext( CPI_ImageContext *info, float time )
{
	/* Get the input information */
	if( cpiInputContext( 0, info, time ) == 0 )
	{
		/* Calculate our output size */
		float t;
		cpiGetFloat( "Tile X", time, &t );
		info->fullX *= t;
		cpiGetFloat( "Tile Y", time, &t );
		info->fullY *= t;

		/* Make sure our image is at least one pixel */
		if( info->fullX <= 0 )
			info->fullX = 1;
		if( info->fullY <= 0 )
			info->fullY = 1;
		return 0;
	}
	else
	{
		cpiError( "Couldn't get input size" );
		return 1;
	}
}

CPIDSOEXPORT int upiCheckInputSize( CPI_ImageContext *output )
{
	/* Since we change the size of the image, any size input is OK */
	return 0;
}

CPIDSOEXPORT int upiCookFull( void )
{
	return 1;
}

/*
 * Figure out what the input size is given the output size.
 * This is basically the reverse of upiGetOutputContext
 * We always cook the input at full size.
 */
CPIDSOEXPORT void upiRegionsNeeded( CPI_ImageContext *output )
{
	float size;
	CPI_ImageContext input;
	cpiInputContext( 0, &input, 0.0 );

	/* Calculate what input size is needed */
	cpiGetFloat("Tile X",output->time, &size);
	input.sizeX = 0;
	if( size > 0 )
		input.sizeX = output->fullX / size;
	cpiGetFloat("Tile Y",output->time, &size);
	input.sizeY = 0;
	if( size > 0 )
		input.sizeY = output->fullY / size;

	/* Make sure we at least get 1 pixel */
	if( input.sizeX <= 0 )
		input.sizeX = 1;
	if( input.sizeY <= 0 )
		input.sizeY = 1;

	input.fullX = input.sizeX;
	input.fullY = input.sizeY;
	input.offsetX = 0;
	input.offsetY = 0;
	input.channels = output->channels;
	input.time = output->time;
	input.input = INPUT_A;
	cpiNeedRegion( &input );
	return;
}

CPI_PluginType upiPluginType (void )
{
	return T_EFFECTS;
}

/*
 * The actual processing function
 * This figures out what type of pixels we have and just
 * calls the templated Tile function at the top
 */
CPIDSOEXPORT int upiProcessImage( CPI_Image *result )
{
	float tx = 0, ty = 0;
	cpiGetFloat( "Tile X", result->info.time, &tx );
	cpiGetFloat( "Tile Y", result->info.time, &ty );
	if( tx <= 0.0 || ty <= 0.0 )
		cpiWarning( "Can't tile 0 times, assuming size of 1 pixel" );

	CPI_Image input;
	if( cpiCookRegion( 0, &input ) != 0 )
		return -1;

	/* Check the output type and call the appropriate tempate */
	switch( result->info.pelType )
	{
		case P_INT8:
		{
			Tile( result, &input, (unsigned char)UCHAR_MAX );
			break;
		}
		case P_INT16:
		{
			Tile( result, &input, (unsigned short)USHRT_MAX );
			break;
		}
		case P_FLOAT32:
		{
			Tile( result, &input, (float)1.0F );
			break;
		}
		default:
			cpiError( "Unknown pixel type" );
			return 1;
	}
	return 0;
}
