/*
 * PROPRIETARY INFORMATION.  This software is proprietary to
 * Silicon Grail Corporation, and is not to be reproduced,
 * transmitted, or disclosed in any way without written permission.
 *
 * Produced by:
 *  Silicon Grail Corporation
 *  710 N. Seward
 *  Hollywood, CA
 *  USA 90038
 *  213-871-9100
 *
 * NAME:  Chalice Plugin Interface
 *
 *   These are the functions available to the plugin
 * author for writing plugins.
 *
 *   This file is best viewed using 4 space tabs.
 *
 */

#ifndef __CHALICE_PLUGIN__
#define __CHALICE_PLUGIN__

#ifdef __cplusplus
extern "C" {
#endif

#include <limits.h>

#define INPUT_GROW UINT_MAX
#define RESULT_SEPARATE UINT_MAX
#define INPUT_A 0
#define INPUT_B 1
#define INPUT_C 2
#define INPUT_D 3

/************************************************************************\
 * Which channels are selected in the select mask
 *  - there is a test function below which will help test which channels
 *    are selected
\************************************************************************/
#define CHANMASK_RED	0x01
#define CHANMASK_GRN	0x02
#define CHANMASK_BLU	0x04
#define CHANMASK_ALP	0x08
#define CHANMASK_OTH	0x10 /* Anything past the alpha channel */

/************************************************************************\
 * Plugin type
 *   Types of plugins
\************************************************************************/
typedef enum
{
	T_FILTER,
	T_COMPOSITE,
	T_TOOL,
	T_IO,
	T_MATTE,
	T_VIDEO,
	T_EFFECTS,
	T_CONVERT,
	T_MOVERS
} CPI_PluginType;

/************************************************************************\
 * Parameter flags 
 *   Flags which parameter can have
\************************************************************************/
enum CPI_ParameterFlag
{
	P_MIN			= 1,
	P_MAX			= 2,
	P_MINMAX		= 3,
	P_ANIMATE		= 4,
	P_HIDDEN	 	= 8
};

/************************************************************************\
 * Pixel type
 *   Type of pels
\************************************************************************/
typedef enum
{
	P_INT8,
	P_INT16,
	P_FLOAT32,
	P_UNKNOWN
} CPI_PelType;

/************************************************************************\
 * Image information 
 *   Used to store various attributes of an image
\************************************************************************/
typedef struct
{
	unsigned int	sizeX;
	unsigned int	sizeY;
	unsigned int	fullX;
	unsigned int	fullY;
	unsigned int	offsetX;
	unsigned int	offsetY;
	unsigned int 	channels;
	float			time;
	CPI_PelType		pelType;
	unsigned int	input;
} CPI_ImageContext;

typedef struct
{
	CPI_ImageContext info;
	void *data;
} CPI_Image;

#ifndef CPIDSOEXPORT
#ifdef WIN32
#define CPIDSOEXPORT __declspec( dllexport )
#else
#define CPIDSOEXPORT
#endif
#endif

#ifndef CPIDSOIMPORT
#ifdef WIN32
#define CPIDSOIMPORT __declspec( dllimport )
#else
#define CPIDSOIMPORT
#endif
#endif

#ifndef DIRECTORY_SEPARATOR
#ifdef WIN32
#define DIRECTORY_SEPARATOR "\\"
#else
#define DIRECTORY_SEPARATOR "/"
#endif
#endif

/* Convert pelType to a byte size */
CPIDSOEXPORT int cpiGetByteSize( CPI_PelType pelType );

/* Get input information */
CPIDSOEXPORT int cpiInputContext( unsigned int		 input,
							   CPI_ImageContext	*info,
							   float time );
CPIDSOEXPORT int cpiNumberOfInputs( void );
CPIDSOEXPORT int cpiInputFrameRange( int i, long *start, long *end );

/* Invalidate the output information */
CPIDSOEXPORT void cpiGetNodeName( char *buffer );

/* Invalidate the output information */
CPIDSOEXPORT void cpiInvalidateOutput( void );

/* Functions to add parameter */
CPIDSOEXPORT void cpiAddTab( char *label );

CPIDSOEXPORT void cpiAddToggle( char	*label,
								int		 def,
								int		 flags );

CPIDSOEXPORT void cpiAddInteger( char	*label,
								 int	 def,
								 int	 min,
								 int	 max,
								 int	 flags );

CPIDSOEXPORT void cpiAddFloat( char		*label,
							   float	 def,
							   float	 min,
							   float	 max,
							   int		 flags );

CPIDSOEXPORT void cpiAddIntVector( char		*label,
								   int		 size,
								   int		*def,
								   int		 min,
								   int		 max,
								   int		 flags );

CPIDSOEXPORT void cpiAddFloatVector( char	*label,
									 int	 size,
									 float	*def,
									 float	 min,
									 float	 max,
									 int	 flags );

CPIDSOEXPORT void cpiAddString( char	*label,
								char	*def,
								int		 flags );

CPIDSOEXPORT void cpiAddMenu( char		*label,
							  int		 def,
							  char		*menu[],
							  int		 flags );

CPIDSOEXPORT void cpiAddChannelMask( char *label, int def, int flags );

CPIDSOEXPORT void cpiAddButton( char *label );

CPIDSOEXPORT void cpiAddFile( char *label, char *def, int flags );

CPIDSOEXPORT void cpiAddPicFile( char *label, char *def, int flags );

/* Get parameter */
CPIDSOEXPORT int cpiGetInteger( char	*param,
								float	 time,
								int		*value );

CPIDSOEXPORT int cpiGetFloat( char		*param,
							  float		 time,
							  float		*value );

CPIDSOEXPORT int cpiGetIntVector( char			*param,
								  float			 time,
								  unsigned int	 vector,
								  int			*value );

CPIDSOEXPORT int cpiGetFloatVector( char			*param,
									float			 time,
									unsigned int	 vector,
									float			*value );

CPIDSOEXPORT int cpiGetString( char		*param,
							   float	 time,
							   char		*buffer );

/* Set parameter */
CPIDSOEXPORT int cpiSetInteger( char	*param,
								float	 time,
								int		 value );

CPIDSOEXPORT int cpiSetFloat( char		*param,
							  float		 time,
							  float		 value );

CPIDSOEXPORT int cpiSetIntVector( char			*param,
								  float			 time,
								  unsigned int	 vector,
								  int			 value );

CPIDSOEXPORT int cpiSetFloatVector( char			*param,
									float			 time,
									unsigned int	 vector,
									float			 value );

CPIDSOEXPORT int cpiSetString( char		*param,
							   float	 time,
							   char		*buffer );

/* Utility functions */
CPIDSOEXPORT int cpiIsChannelSelected( int			 select,
									   unsigned int	 channel );

/* Disable/enable parameter */
CPIDSOEXPORT void cpiEnableParameter( char *name, int onoff );

/* Get various paths in chalice */
#define CPI_STRING_LENGTH 2048
#define CPI_PATH_LENGTH 2048

/* Get various chalice paths */
CPIDSOEXPORT void cpiChalicePath( char *path );
CPIDSOEXPORT void cpiPluginPath( char *path );
CPIDSOEXPORT void cpiTmpPath( char *path );
CPIDSOEXPORT void cpiUserPath( char *path );

/* Pass a message to chalice */
CPIDSOEXPORT void cpiError( char *fmt, ... );
CPIDSOEXPORT void cpiWarning( char *fmt, ... );

/* Doing various things with regions */
CPIDSOEXPORT void cpiNeedRegion( CPI_ImageContext *input );
CPIDSOEXPORT int  cpiCookRegion( int i, CPI_Image *img );
CPIDSOEXPORT void cpiFreeRegion( int i, CPI_Image *img );

/* Time/Frame conversions */
CPIDSOEXPORT int		cpiGetFrame( float time );
CPIDSOEXPORT float		cpiGetTime( int frame );
CPIDSOEXPORT int		cpiSetFrameRange( long start, long end );

/* Multi processing function */
CPIDSOEXPORT int		cpiNumProcessors( void );
CPIDSOEXPORT void		cpiMultiProcess( void (*func)( int, int, void * ), void *arg );

/* String functions */
CPIDSOEXPORT void		cpiExpandString( char *string, float t );

/* User functions */
CPIDSOEXPORT int				 upiInitialize( int workflag );
CPIDSOEXPORT char				*upiAuthorName( void );
CPIDSOEXPORT char				*upiPluginName( void );
CPIDSOEXPORT char				*upiCopyrightNotice( void );
CPIDSOEXPORT char				*upiPluginVersion( void );
CPIDSOEXPORT CPI_PluginType		 upiPluginType( void );
CPIDSOEXPORT void				 upiNumberOfInputs( int *min, int *max );
CPIDSOEXPORT void				 upiCreateParameters( void );
CPIDSOEXPORT void				 upiRegionsNeeded( CPI_ImageContext *output );
CPIDSOEXPORT int				 upiProcessImage( CPI_Image *result );
CPIDSOEXPORT char				*upiInputLabel( unsigned int input );
CPIDSOEXPORT char				*upiOutputLabel( void );
CPIDSOEXPORT void				 upiParameterChanged( char *name );
CPIDSOEXPORT int				 upiCheckInputSize( CPI_ImageContext *result );
CPIDSOEXPORT int				 upiCheckInputDepth( CPI_ImageContext *result );
CPIDSOEXPORT int				 upiResultInput( CPI_ImageContext *result );
CPIDSOEXPORT int				 upiModifyInput( unsigned int input );
CPIDSOEXPORT char				*upiMessageInfo( float time );
CPIDSOEXPORT int				 upiOutputContext( CPI_ImageContext *output, float time );
CPIDSOEXPORT void				 upiInputChanged( void );
CPIDSOEXPORT int				 upiCookFull( void );
CPIDSOEXPORT void				 upiNodeCreated( void );
CPIDSOEXPORT void				 upiNodeDeleted( void );
CPIDSOEXPORT void				 upiNameChanged( char *, char * );

#ifdef __cplusplus
}
#endif

#endif

