#! /usr/bin/perl
print <<EOF ;

#######################################################################
#######################################################################
#####                                                             #####
#####                                                             #####
#####      Perl script to create an alien star catalog for use    #####
#####      with the "Stars" plugIn for Chalice                    #####
#####                                                             #####
#####      Written by Dirk P. Runge.                              #####
#####      Copyright 1998, Silicon Grail Inc.                     #####
#####                                                             #####
#####                                                             #####
#######################################################################
#######################################################################

EOF


#######################################################################
#####                                                             #####
#####      Configuration: Change the following lines to suit      #####
#####      your application.                                      #####
#####                                                             #####
#######################################################################

############################
#####                  #####
#####      Colors      #####
#####                  #####
############################

# This table ( okay, list ) describes the colors of the stars in the catalog
# The first number ( in quotes ) is the relative distribution of the color.
# The next three numbers are the relative RGB values.

# All of these numbers will be normalised, and may be expressed as either
# positive integers, or floating point values at your discretion.

@colors = (	'43961', 1.0, 1.0, 1.0, 	# Others ( white )
		   '94',  58, 103, 255, 	# O
		'15145', 114, 207, 255, 	# B
		'45637', 128, 218, 253, 	# A
		'43830', 164, 255, 255, 	# F
		'41237', 206, 255, 255, 	# G
		'64123', 233, 255, 190, 	# K
		 '4969', 254, 249, 102) ;	# M

############################
#####                  #####
#####    Brightness    #####
#####                  #####
############################

# This function call describes the brightness of the star being generated
# The star number ( starting with 1 ) is passed to the function, and a 
# return value of 1 or less is expected. This value will be used to linearly
# scale the brightness of the rendered star.
# NOTE:
#      * &brightness( 1.0 ) must return 1.0 ( the first star is 1.0 )
#      * The function must decrease ( or stay the same ) for each star.

# Initialization code: ( Gets executed once. )
$exponent = -1.8 / log(10) ;

# Function call:
sub brightness {
	local( $star ) = $_[0] ;

	return $star ** $exponent ;
}


############################
#####                  #####
#####    Postition     #####
#####                  #####
############################

# This function call returns the X, Y, Z position of the star being generated
# The star number ( starting with 1 ) is passed to the function.
# The resulting X, Y, Z values will be normalized.
# DO NOT return a 0, 0, 0 vector.

# Initialization code: ( Gets executed once. )
	#none ;

# Function call:
sub position {
	$y   = (2 * rand) - 1 ;
	$rad = sqrt(1 - ($y ** 2)) ;
	$ra  = 2 * $pi * rand ;
	$x = ($rad * sin($ra)) ;
	$z = ($rad * cos($ra)) ;
	return ( $x, $y, $z ) ;
}

############################
#####                  #####
#####      Format      #####
#####                  #####
############################

# This script understands floating point numbers in both MIPS and Intel formats
# As of this writing, the Stars plugIn for chalice will understand both formats
# and be able to use them interchangeably. Because of this, you probably won't
# have to change anything here.
#
# If you wish to select a particualar format, uncomment the associated line
# below. If by chance, you are using a CPU with a different format, and this
# is *exteremely* unlikely, uncomment the "other" line.
# If all lines are commented this script will automatically determine the format
# of the machine running this script and build the file to that format.

# $cpu = 'MIPS'  ;
# $cpu = 'Intel' ;
# $cpu = 'other' ;

#######################################################################
#####                                                             #####
#####      User interface: You shouldn't have to change anything  #####
#####      after this point.                                      #####
#####                                                             #####
#######################################################################

package starGen ;
&init ;  # Verify configuration first

RESTART: print <<EOF ;
Please enter a name for the "location" of the starfield that your are
generating. This name is used mainly for reference. Check the script,
or your shot list to get this information.
NOTE:
If your shot is anywhere near the Earth, like somewhere in this solar
system, you probably want to use the included earthStars.sdb file instead
of generating a random catalog.
EOF

&getUI( "Location Name", $location, '/\w/',
 'Location must contain "word" characters (A-Z, 0-9)') ;


$size = '5.5 Megabytes' unless $size ;
# Choose the number of stars, or the file size
print <<EOF ;
Please select a maximum number of stars, or a maximum file size as follows:
For a fixed number of stars enter: \$count Stars
For a fixed file size enter:       \$size Bytes
                                   \$size Kilobytes
                                   \$size Megabytes
                                   \$size Gigabytes
For example: If you want a 20 Kilobyte file, enter 20k.
Remeber that the entire catalog is loaded into memory when the plugin executes
EOF

&getUI( "File size", $size, '/^\d+\.?\d*\s*[S|B|K|M|G]/i',
 'I did not understand your entry, please try again.') ;

# Generate the random number seed.
print <<EOF ;
Please select a random number seed value. Each seed value will produce a
unique star catalog.
Note:
There is no guarantee that the same seed will always produce the same catalog.
Variations in platforms, operating system, perl and code versions may
cause different catalogs to be generated from the same seed number.
EOF
unless ($seed) {
	$seed = 0 ;
	foreach $element (unpack("C*",$location)) {
		$seed += $element ;
	}
	$seed = ( $seed % 65535 ) + 1;
}
&getUI( "Seed value", $seed, '$_ > 0', "Seed value must be greater than 0") ;


# Choose the filename
REFILE: print <<EOF ;
Please select the filename to save this star catalog as.
Note that if the file name doesn't end in .sdb it will be appended
EOF

unless ($filename) {
	foreach $element (split(/\s+/,$location)) {
		$element =~ s/\W//g ;
		next unless $element ;
		$element =~ tr/A-Z/a-z/ ;
		$filename = $element if length($element) > length($filename) ;
	}
	$filename .= '.sdb' ;
}
&getUI( "Filename", $filename ) ;
$filename .= '.sdb' unless $filename =~ /\.sdb$/i ;

# Verify the info
print <<EOF ;
Please verify these values:

Location name: $location
File size:     $size
Seed value:    $seed
Filename:      $filename

EOF

$ok = 'Yes' ;
&getUI( "Everything Okay", $ok, '/^(Y|N)/i', "Please enter Yes or No" ) ;
goto RESTART if $ok =~ /n/i ;

# Test permissions / existence of the file
if ( -e  $filename ) {
	$ftest = 'No' ;
	&getUI( "File exists overwrite", $ftest, '/^(Y|N)/i',
	 "Please enter Yes or No" ) ;
	goto REFILE if $ftest =~ /n/i ;
}

# open the file
unless ( open(OUT, ">$filename") ) {
	print "Unable to open file, $filename, $!\n" ;
	goto REFILE ;
}

sub getUI { # (Prompt string, value, condition, failure notice)
	local($prompt,$old,$condition,$failure)  = @_ ;
	$condition = 1 if @_ < 3 ;
	print $prompt,': ' ;
	if ( $old ) {
		print '(', $old, ') ' ;
	}
	$_ = <STDIN> ;
	s/\s+$// ;
	s/^\s+// ;
	$_ = $old unless /\S/ ;
	until( eval($condition) ) {
		print $failure, "\n", $prompt, ': ' ;
		if ( $old ) {
			print '(', $old, ') ' ;
		}
		$_ = <STDIN> ;
		s/\s+$// ;
		s/^\s+// ;
		$_ = $old unless /\S/ ;
	}
	$_[1] = $_ ;
	print "\n" ;
}

#######################################################################
#####                                                             #####
#####      Calculation: Don't change anything after this point    #####
#####      unless you are SURE you know what you are doing.       #####
#####                                                             #####
#######################################################################

#################################################
#                                               #
#  Verify configuration, and set up color list  #
#                                               #
#################################################
sub init { # Verify configuration, and set up color list
	# Copy colors
	@colors = @main'colors ;

	# Do the colors first
	$sum = 0 ;
	while (@colors) {
		$d = shift(@colors) ;
die( "CONFIGURATION ERROR: Color distribution cannot be less than 0\n" ) 
		if $d < 0.0 ;

		$r = shift(@colors) ; 
die( "CONFIGURATION ERROR: Red value cannot be less than 0\n" ) 
		if $r < 0.0 ;

		$g = shift(@colors) ; 
die( "CONFIGURATION ERROR: Green value cannot be less than 0\n" ) 
		if $g < 0.0 ;

		$b = shift(@colors) ; 
die( "CONFIGURATION ERROR: Blue value cannot be less than 0\n" ) 
		if $b < 0.0 ;

		$factor = ( $r + $g + $b ) ;
die( "CONFIGURATION ERROR: Color values cannot sum to 0\n" ) 
		if $factor == 0.0 ;

		$factor = 3 / $factor ;
		push(@R,$r * $factor) ; push(@G,$g * $factor) ;
		$sum += $d ;
		push(@D,$sum) ;
	}
die( "CONFIGURATION ERROR: Color distribution cannot sum to 0\n")
	if $sum == 0.0 ;
	foreach $d (@D) {
		$d /= $sum ;
	}

	# initialize the color look up table
	for ($index = 0 ; $index < 256 ; $index++ ) {
		if ($R[$index]) {
			$red[$index]   = $R[$index] ;
			$green[$index] = $G[$index] ;
		} else {
			$red[$index]   = 1 ;
			$green[$index] = 1 ;
		}
	}

	# Define pi
	$pi = atan2(1,1) * 4 ;
	$main'pi = $pi ;

	# Check that brightness decreases
die( "CONFIGURATION ERROR: &brightness( 1 ) != 1.0\n") 
	if &main'brightness( 1 ) != 1.0 ;
die( "CONFIGURATION ERROR: &brightness( $count ) increases with $count\n")
	if &main'brightness( 1 ) < &main'brightness( 2 ) ;

	# Check the processor
	@cpu = (6.02E23, "\x66\xfe\xf4\xf9", 
		5.0002,  "\x40\xa0\x01\xa3",
		1.0,     "\x3f\x80\x00\x00",
		0.0,     "\x00\x00\x00\x00") ;

	$float = $cpu[0] ;
	$code  = $cpu[1] ;
	$byteOrder =  0 ; # unknown byte order
	# test if MIPS order
	$byteOrder =  1 if $code eq pack("f",$float) ;
	# test if INTEL order
	$byteOrder = -1 if reverse($code) eq pack("f",$float) ;
	# Bytes to be passed to the plugin to test
	# the float format against the native one
	$testBytes = '' ;
	while ( @cpu ) {
		$float = shift(@cpu) ;
		$code  = shift(@cpu) ;
		$bytes = pack("f",$float) ;
		$testBytes .= $bytes ;
		if ($byteOrder == 1) {
			$byteOrder = 0 if $code ne $bytes ; ;
		} else {
			$byteOrder = 0 if reverse($code) ne $bytes ;
		}
	}

	$cpu = $main'cpu ;
	if ($byteOrder == 1) {
		# MIPS order detected
		$swap = 1 if $main'cpu =~ /intel/i ;
		$cpu  = 'MIPS' unless $cpu ;
	} elsif ($byteOrder == -1) {
		# INTEL order detected
		$swap = 1 if $main'cpu =~ /mips/i ;
		$cpu  = 'Intel' unless $cpu ;
	} else {
die( "CONFIGURATION ERROR: Unknown CPU byte order.
Uncomment the \$cpu = 'other' ; line") unless $main'cpu =~ /other/i ;
	}
}

# Unbuffer STDOUT
select((select(STDOUT), $| = 1)[0]) ;

# Start writing to the output file
$header1  = "Stars " ; # 6 bytes required
$header1 .= 1 ; # major revision
$header1 .= '.' ;
$header1 .= 1 ; # catalog number
$header1 .= '.' ;
$header1 .= 0 ; # minor revision
$header1 .= " Sorted \n" ;

# leave a space for the starcount
$header2  = " stars in the catalog.\n" ;
$header2 .= "Random star catalog located at $location.\n" ;
$header2 .= "$cpu float format\n\0" ;

# Design the final "Credits"
$bar =
"+--------------------------------------------------------------------------+
" ;
$title2 = $line = 
"|                                                                          |
" ;

$title  = "Randomly generated star catalog $seed for $location" ;
$title  = substr( $title, 0, length( $title2 ) - 5 ) ;
substr( $title2, 3, length( $title ) ) = $title ;

$title = $bar . $line . $title2 . $line . $bar ;

# Parse the starcount field
die("Assertion failed") unless $size  =~ /^(\d+\.?\d*)\s*([S|B|K|M|G])/i ;
$count =  $1 ;
$type  =  $2 ;
$type  =~ tr/a-z/A-Z/ ;
unless ( $type eq 'S') {
	# Memory size, convert to bytes
	     if ( $type eq 'G' ) {
		$count *= 1024 * 1024 * 1024 ;
	} elsif ( $type eq 'M' ) {
		$count *= 1024 * 1024 ;
	} elsif ( $type eq 'K' ) {
		$count *= 1024 ;
	} elsif ( $type eq 'B' ) {
		# do nothing
	} else {
		die("Assertion failed, unknown size type") ;
	}

	# Remove overhead from memory count
	$count -= 10 ; # Leave 10 bytes for the star count
	$count -= length( $header1 )   ; # header goes
	$count -= length( $header2 )   ; # header goes
	$count -= 512 * 4 ; # color look up table
	$count -= length( $testBytes ) ; # those testbytes
	$count -= 1 ; # And the checkpoint ;
	$count -= 4 ; # And the literal EOF ;
	$count -= length( $title ) ; # The footer

	# See how many stars we can fit
	$count = int( $count / 17 ) ; # 17 bytes per star 
} else {
	$count += 1 ; # include the bogus star
}

# Print the header and testbytes
print OUT $header1, $count, $header2;

binmode( OUT ) ; # It's binary from this point onward.

print OUT $testBytes ;

$count -= 1 ; # Don't include the bogus star at this time

# Write to the Color Look Up Table
for ($index = 0 ; $index < 256 ; $index++ ) {
	if ($swap) {
		print OUT scalar( reverse( pack( "f f", $green[$index],
		 $red[$index] ) ) ) ;
	} else {
		print OUT pack( "f f", $red[$index], $green[$index] ) ;
	}
}

# Write checkpoint. Get it? Checkpoint! Ha! Ha!
print OUT "." ;

print "Writing $count stars to the file $filename\n" ;

# Be nice to the user, and show a progress bar.
print " 0% 10% 20% 30% 40% 50% 60% 70% 80% 90% 100%\n" ;
print " |   |   |   |   |   |   |   |   |   |    | \n" ;
print " *" ;
$increment = $count / 40 ;
@increment = 1..39 ;
foreach $element (@increment) {
	$element *= $increment ;
	$element = int($element + .5) ;
}
push( @increment, $count ) ;
$thisStar  = 1 ;
$count    -= 1 ; # Do not include the bogus star in the count
while ( @increment ) {
	$increment = shift( @increment ) ;
	while ( $thisStar <= $increment ) {
		$mag = &main'brightness( $thisStar ) ;
		( $x, $y, $z ) = &main'position( $thisStar ) ;
		$color = rand ;
		$COLOR = 0 ;
		while ($color > $D[$COLOR]) {
			$COLOR ++ ;
		}
		if ($swap) {
			$starDat .= scalar( reverse(
			 pack( "ffff", $z, $y, $x, $mag ))) ;
		} else {
			$starDat .= pack( "ffff", $mag, $x, $y, $z ) ;
		}
		$starCol .= pack("C",$COLOR) ;
		$thisStar ++ ;
	}
	print "*" ;
}
print "*\n\n" ;

# Last star is a bogus '0' star
if ($swap) {
	$starDat .= scalar( reverse( pack( "ffff", 0.0, 0.0, 0.0, 0.0 ))) ;
} else {
	$starDat .= pack( "ffff", 0.0, 0.0, 0.0, 0.0 ) ;
}
$starCol .= pack( "C", 0.0 ) ;

# Print the stored data to the files
print OUT $starCol, $starDat, "EOF\0" ;

print OUT $title ;

__END__
